"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.arrayhash

Provides the ArrayHash class. An ArrayHash allocates indices on an array
such that an ordering is maintained. This is accomplished by a bidirectional
linear hash that preserves the ordering parameter. 
"""
from __future__ import division

class RenderSlots(list):
    """Object containing render slots; will eventually support the
    reassignment of render slots"""

class ArrayHash(object):
    """The ArrayHash object allocates indices on an array such that
    a given ordering is preserved; note that the array may contain
    zero'd elements which are blanks. For optimal performance,
    the distribution of the elements ordering should be uniform
    (although doesn't have to be continuous)"""
    def __init__(self, size, start, end, groupSize=1):
        """Initialise the ArrayHash
        Arguments:
          size: The size of the array to manage
          start: The beginning of the ordering
          end: The end of the ordering
          groupSize: The approximate minimum size
          of the groups of elements to be allocate. Setting
          this too high will result in underallocation, too low
          may lead to less than optimal performance."""
        self.lastIndex = size - 1
        self.start = start
        self.mult = size / (end - start)  
        self.allocs = {}
        self.positionJumps = {}
        self.groupSize = groupSize
        
    def __alloc(self, number, position, maxIndx, minIndx):
        """Internal version of alloc, exposing the maxIndx and minIndx
        parameters which are used for a loopback search"""
        base = (position - self.start) * self.mult
        indx = int(base - (base % self.groupSize))
        forwardIndx, backwardIndx = self.positionJumps.get(position, 
                                                           (indx, indx-1))
        forwardIndx = max(forwardIndx, 0)
        backwardIndx = min(backwardIndx, self.lastIndex)
        oForward, oBackward = forwardIndx, backwardIndx
        searchForward, searchBackward = True, True
        found = False
        foundIndx = None
        while found is False and (searchForward or searchBackward):
            if searchForward:
                if (forwardIndx + number > maxIndx or 
                any((self.allocs[forwardIndx+x] > position 
                for x in xrange(number) if forwardIndx+x in self.allocs))):
                    searchForward = False
                elif not any(forwardIndx + x in self.allocs 
                for x in xrange(number)):
                    found = True
                    foundIndx = forwardIndx
                    forwardIndx += number
                else:
                    forwardIndx += self.groupSize
            elif searchBackward and not found:
                if (backwardIndx - number < minIndx or 
                any((self.allocs[backwardIndx-x] < position 
                for x in xrange(number) if backwardIndx-x in self.allocs))):
                    searchBackward = False
                elif not any(backwardIndx - x in self.allocs 
                for x in xrange(number)):
                    found = True
                    backwardIndx -= number
                    foundIndx = backwardIndx + 1
                else:
                    backwardIndx -= self.groupSize
        if not found:
            if position in self.positionJumps:
                # If we skipped any positions, they may have been deallocated,
                # so go back and check
                del self.positionJumps[position] 
                return self.__alloc(number, position, maxIndx=oForward, 
                                    minIndx=oBackward)
            else:
                return False
        else:
            for x in xrange(number):
                self.allocs[foundIndx + x] = position
                
            self.positionJumps[position] = forwardIndx, backwardIndx
            return RenderSlots((foundIndx + x for x in xrange(number)))
        
    def alloc(self, number, position):
        """Allocate indices on the array.
        Argument:
          number: The number of indicies required.
          position: Desired relative position in the array to other
                    elements. All elements before have lower position,
                    elements after have higher position"""
        return self.__alloc(number, position, self.lastIndex, 0)
    
    def release(self, indices):
        """Release allocated indices"""
        for indx in indices:
            del self.allocs[indx]

def stress(deallocProb):
    """A simple stress tester"""
    import random, time
    while True:
        arraySize = 10000
        test = ArrayHash(arraySize, 0, 10, groupSize=4)
        going = True
        noAllocd = 0
        noDeAllocd = 0
        allocs = []
        while going:
            start = time.time()
            if random.random() > deallocProb:
                pos = random.randint(0, 9)
                res = test.alloc(4, pos)
                if res is False:
                    timeTaken = time.time() - start
                    print ('Allocated %i of %i with %i deallocs in %f' % 
                           (noAllocd, arraySize, noDeAllocd, timeTaken))
                    going = False
                else:
                    noAllocd += len(res)
                    allocs.append(res)
            else:
                if len(allocs) > 0:
                    indx = random.randint(0, len(allocs)-1)
                    toRelease = allocs.pop(indx)
                    test.release(toRelease)
                    noDeAllocd += len(toRelease)
            

if __name__ == '__main__':
    stress(-1)
