"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.examples.criticalmass

Provides a simple "critical mass" or "contagion" type game
"""

from exampleskeleton import ExampleSkeleton, pyg3
import random
import os.path
from math import sqrt
from pygame.locals import MOUSEBUTTONUP

class Ball(object): 
    """A Ball in the level"""
    def __init__(self, level):
        """Initialise the ball"""
        self.xpos = random.randint(0, level.xpos)
        self.ypos = random.randint(0, level.ypos)
        self.vecx = random.randint(-level.maxspd, level.maxspd)
        self.vecy = random.randint(-level.maxspd, level.maxspd)
        self.color = (random.randint(0, 255), random.randint(0, 255), 
                      random.randint(0, 255), 128)
        self.sprite = pyg3.Sprite(image=level.ballImg, color=self.color, 
                                  position=(self.xpos, self.ypos), scale=0.2, 
                                  offsets=pyg3.centerOffsets)
        self.radius = 10
        self.scale = 0.2
        self.exploding = False
        self.explosion = 0
        
    def tick(self, level):
        """Perform a single tick"""
        self.xpos += self.vecx
        self.ypos += self.vecy
        
        if self.xpos < 0:
            self.xpos = -self.xpos
            self.vecx = -self.vecx
        elif self.xpos > level.xpos:
            self.xpos = level.xpos - (self.xpos - level.xpos)
            self.vecx = -self.vecx
        if self.ypos < 0:
            self.ypos = -self.ypos
            self.vecy = -self.vecy
        elif self.ypos > level.ypos:
            self.ypos = level.ypos - (self.ypos - level.ypos)
            self.vecy = -self.vecy    
        
        self.sprite.setPosition((self.xpos, self.ypos))
    
        # Check to see if we should be exploding
        for exploder in level.exploders:
            dx = abs(self.xpos - exploder.xpos)
            dy = abs(self.ypos - exploder.ypos)
            if dx < exploder.radius and dy < exploder.radius:
                dist = sqrt((dx * dx) + (dy * dy))
                if dist < exploder.radius:
                    # we explode
                    self.exploding = exploder.exploding + 1
                    self.sprite.setImage(level.bangImg)


class Bang(object):
    """An exploding ball in the level"""
    def __init__(self, level, pos, sprite=None):
        """Initialise the bang"""
        self.xpos = pos[0]
        self.ypos = pos[1]
        self.vecx, self.vecy = 0, 0
        self.color = (255, 255, 255, 255)
        if sprite is None:
            sprite = pyg3.Sprite(image=level.bangImg, color=self.color, 
                                  position=(self.xpos, self.ypos), scale=0.2,
                                  offsets=pyg3.centerOffsets)
        self.sprite = sprite
        self.radius = 10
        self.scale = 0.2
        self.exploding = 0
        self.explosion = 0
        
    def tick(self):
        """Update the bang"""
        if self.explosion < 10:
            self.radius *= 1.1
            self.scale *= 1.1
            self.sprite.setScale(self.scale)
        if self.explosion > 100:
            self.radius /= 1.1
            self.scale /= 1.1
            self.sprite.setScale(self.scale)
        if self.radius < 2:
            self.sprite.setVisibility(False)
            self.exploding = False
        self.explosion += 1        
       
class Level(object):
    """A level, currently not actually used much"""
    def __init__(self, number, size):
        """Initialise the level"""
        self.xpos, self.ypos = size
        self.maxspd = int(number * 0.4 + 1)
        self.balls = int((number * number) * 0.6 + (2 * number))
        self.required = (number * 0.8) * 0.8 * self.balls
        self.ballSize = max(10 - (number / 5), 1)
        self.ballImg = pyg3.loadImage(os.path.join('data', 'ball.png'))
        self.bangImg = pyg3.loadImage(os.path.join('data', 'bang.png'))
        self.balls = set([Ball(self) for _ in xrange(self.balls)])
        self.exploders = []
        self.bangPlaced = False
        self.finished = False
        
    def placeBang(self, pos):
        """Function to place the initial bang on click"""
        if not self.bangPlaced:
            self.bangPlaced = True
            self.exploders.append(Bang(self, pos))
        
    def tick(self):
        """Perform a single tick"""
        removes = []
        for ball in self.balls:
            ball.tick(self)
            if ball.exploding is not False:
                removes.append(ball)
                self.exploders.append(
                    Bang(self, (ball.xpos, ball.ypos), ball.sprite))
        for rball in removes:
            self.balls.remove(rball)
        for exploder in self.exploders:
            exploder.tick()
            if exploder.exploding is False:
                self.exploders.remove(exploder)
        if self.bangPlaced and len(self.exploders) == 0:
            self.finished = True

class CriticalMass(ExampleSkeleton):
    """The critical mass game"""
    def __init__(self):
        """Initialise the window"""
        self.size = (400, 400)
        super(CriticalMass, self).__init__('CriticalMass', self.size)
        
    def init(self):
        """Sets up the level"""
        self.level = Level(10, self.size)
        
    def eventHandler(self, event):
        """Dispatch events"""
        if event.type == MOUSEBUTTONUP:
            self.level.placeBang(event.pos)
        
    def loop(self, dt):
        """Perform the game loop"""
        self.level.tick()
        if self.level.finished:
            self.going = False

if __name__ == '__main__':
    x = CriticalMass()
    x.run()
    
