"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3).

pygl3display.examples.platformer

This file implements a platforming demo, which may become a full game in the
future.
"""
from __future__ import division

#Import Modules
import os, os.path, random, sys, operator
from collections import defaultdict
import pygame
from pygame.locals import (KEYUP, KEYDOWN, K_LEFT, K_UP, K_RIGHT, QUIT, 
                           K_ESCAPE)
from math import copysign
# Magic so that examples work if trying from an extracted archive
try:
    import pygl3display
except ImportError:
    sys.path.append(os.path.join('..', '..'))
    import pygl3display
pygl3display.configure('pygame')
from pygl3display import (loadImage, Sprite, makeAdjacentSprites, 
                          TwoDVertexCamera, topLeftOffsets, SpriteCommon,
                          TwoDFragmentCamera, GLCommon)



DEBUG = False

class BroadcastSet(set):
    """A set of functions which should all have the same function signature.
    When broadcast is called, all functions inside are called"""
    def broadcast(self, *args, **kwargs):
        """Broadcast to all functions in the set"""
        for func in self: 
            func(*args, **kwargs)
        
class Broadcaster(dict):
    """A broadcasting system in native types. Usage:
    broadcaster[signal].add(function): add function to listening to signal
    broadcaster[signal].broadcast(args): broadcast to everything listening
                                         to signal"""
    def __missing__(self, key):
        """Automatically create the broadcast sets for new signals"""
        self[key] = BroadcastSet()
        return self[key]


class ScheduledTask(object):
    """A deferred call to a function"""
    def __init__(self, function, args=None, kwargs=None):
        """Initialise the deferred task.
        Args:
         function: function to call
         args: a list of arguments to the function
         kwargs: a dict of kwargs to the function"""
        if args is None: args = []
        if kwargs is None: kwargs = {}
        self.function = function
        self.args = args
        self.kwargs = kwargs
        
    def __call__(self):
        """Syntactic sugar"""
        self.function(*self.args, **self.kwargs)

class Scheduler(dict):
    """Schedules functions so that they can happen every n tick
    events"""
    def __init__(self, current, mod, broadcastSystem, 
                 *args, **kwargs):
        """Set up the scheduler.
        Args:
          current: Current tick
          mod: Value for time to be modulo
          broadcastSystem: The broadcastSystem to use"""
        super(Scheduler, self).__init__(*args, **kwargs)
        self.current = current
        self.mod = mod
        broadcastSystem['tick'].add(self.tick)
    
    def __hash__(self):
        """We need hashing to be able to receive a broadcast, but 
        dicts in general are not hashable. As we don't care about the nice
        properties of a hash in this case, we can just take the hash function of
        object."""
        return object.__hash__(self)
        
    def __getitem__(self, key):
        """Get the tasks at a given tick relative to the current
        time (i.e. scheduler[3] = 3 ticks ahead"""
        if isinstance(key, (int, float)):
            key = ((key + self.current) % self.mod,)
        return super(Scheduler, self).__getitem__(key)
            
    def __missing__(self, key):
        """Create a missing set of tasks at a given time"""
        if isinstance(key, tuple):
            self[key] = set()
            return self[key]
        else:
            raise KeyError(str(key) + ' not in scheduler')
        
    def tick(self, dt):
        """Perform a tick"""
        old = int(self.current)
        self.current += dt
        mod = self.mod
        for key in xrange(old, int(self.current)):
            key = (key % mod,)
            if key in self:
                for task in self[key]:
                    task()
                self[key].clear()
        self.current %= mod


class ResourceStore(object):
    """Stores resources loaded from disk"""
    def __init__(self, loader, prefix, suffix):
        """Initialise the ResourceStore
        Args:
          loader: An unary function with one argument, the resource to load
          prefix: A prefix to names to turn them into filenames
                  (e.g. resource directory name)
          suffix: A suffix to names to turn them into filenames
                  (e.g. file extension)
        """
        self.loaded = {}
        self.loader = loader
        self.prefix = prefix
        self.suffix = suffix
        
    def __call__(self, resource):
        """Get a resource; if not already loaded, load it from disk by
        prepending the prefix and appending the suffix"""
        if resource not in self.loaded:
            resourceFile = os.path.join(self.prefix, resource + self.suffix)
            self.loaded[resource] = self.loader(resourceFile)
        return self.loaded[resource]
        
                
class World(object):
    """Holds and update the world state"""
    def __init__(self, ground, dudes, broadcastSystem):
        """Initialise the world state"""
        self.floor = ground.outlineHeights
        self.groundBaddies = []
        self.aerialBaddies = []
        self.dudes = dudes
        self.collider = {}
        self.gravity = -0.5
    
        self.friction = 1 - 0.2
        self.air = 1 - 0.1
        self.lvlheight = 400
        broadcastSystem['tick'].add(self.tick)
        
    def groundMove(self, thing, dt, grav, fric, air):
        """Perform a ground-based movement on a thing"""
        if not thing.resting:
            thing.yvel += thing.yacc * dt
            thing.yvel += grav
            resist = air
        else:
            thing.yvel = 0
            resist = fric
        thing.xvel += thing.xacc * dt
        if thing.yvel > 0:
            thing.yvel = min(thing.yvel, thing.yterm)
        else:
            thing.yvel = max(thing.yvel, -thing.yterm)
        thing.xvel *= resist
        absxvel = abs(thing.xvel)
        if absxvel > thing.xterm: 
            thing.xvel = copysign(thing.xterm, thing.xvel)
        elif absxvel < 0.1:
            thing.xvel = 0.0
        xpos = int(thing.xpos + thing.xvel*dt)
        xpos2 = xpos + thing.width
        maxAdj = abs(thing.xpos - xpos)
        climbHeight = thing.climbHeight if thing.resting else 0
        if thing.xvel > 0:
            while self.floor[xpos2] - climbHeight > thing.ypos and maxAdj:
                xpos2 -= 1
                xpos -= 1
                maxAdj -= 1
        elif thing.xvel < 0:
            while self.floor[xpos] - climbHeight > thing.ypos and maxAdj:
                xpos += 1
                xpos2 += 1
                maxAdj -= 1
        thing.xpos = xpos
        floory = max(self.floor[xpos], self.floor[xpos2])
        thing.ypos = int(max(thing.ypos + thing.yvel*dt, floory))
        thing.resting = thing.ypos == floory
            
        thing.camera.setCameraPosition(
            (thing.xpos, self.lvlheight - thing.ypos - thing.height))
        if thing.ypos == 0:
            thing.kill()
        
    def tick(self, dt):
        """Perform a single clock tick"""
        gravDist = self.gravity * dt
        fric = self.friction
        air = self.air
        self.collider.clear()
        groundmove = self.groundMove
        dels = []
        for baddie in self.groundBaddies:
            groundmove(baddie, dt, gravDist, fric, air)
        for dude in self.dudes:
            groundmove(dude, dt, gravDist, fric, air)
            if dude.dead:
                dels.append(dude)
        for dude in dels:
            self.dudes.remove(dude)

class GroundMover(object):
    """Must have the following attributes:
       xpos, ypos, xvel, yvel, xacc, yacc
    """
    
    
class CameraManager(object):
    """Causes cameras to follow something"""
    def __init__(self, following, cameras, zoom, windowSize, broadcastSystem, 
                 limits=None):
        """Initialise the camera manager"""
        if limits is None: limits = {}
        self.cameras = cameras
        self.following = following
        for camera in self.cameras:
            camera.setCameraScale(zoom)
        self.zoom = float(zoom)
        self.windowSize = windowSize
        self.middleArea = (self.windowSize[0]*0.2, self.windowSize[1]*0.3,
                           self.windowSize[0]*0.8, self.windowSize[1]*0.7)
        self.offsetY = -self.windowSize[1]*(zoom - 1)
        self.current = {}
        self.current['xpos'] = 0
        self.current['ypos'] = 0
        self.limits = limits
        for limit, default in (('xpos', (0, 1000*zoom)), 
                               ('ypos', (0, 1000*zoom))): 
            if limit not in self.limits:
                self.limits[limit] = default
        broadcastSystem['tick'].add(self.tick)
        self.cameraMoveConditions = (
            (operator.lt, 'ypos', self.middleArea[1], -1), 
            (operator.gt, 'ypos', self.middleArea[3], 1),
            (operator.lt, 'xpos', self.middleArea[0], -1), 
            (operator.gt, 'xpos', self.middleArea[2], 1))
            
    def tick(self, dt):
        """Perform an update"""
        followx, followy = self.following.xpos, self.following.ypos
        display = {}
        display['xpos'] = (followx * self.zoom) - self.current['xpos']
        display['ypos'] = (followy * self.zoom) - self.current['ypos']
        for test, direction, boundVal, mult in self.cameraMoveConditions:
            if test(display[direction], boundVal):
                delta = abs(display[direction] - boundVal) 
                self.current[direction] += (delta * mult)
                lowerlimit, upperlimit = self.limits[direction]
                if self.current[direction] < lowerlimit:
                    self.current[direction] = lowerlimit
                elif self.current[direction] > upperlimit:
                    self.current[direction] = upperlimit
                self.dirty = True
        if self.dirty:
            self.dirty = False
            newPos = (-self.current['xpos'], self.current['ypos']+self.offsetY)
            for camera in self.cameras:
                camera.setCameraPosition(newPos)
        
    
class Dude(GroundMover):
    """A dude controlled by the player"""
    def __init__(self, groundCamera, scheduler, broadcastSystem, imageStore):
        """Initialise the dude"""
        images = [imageStore(x) 
                  for x in ['leg', 'leg', 'body', 'arm', 'arm', 'head']]
        self.sprites = makeAdjacentSprites(images, 0)
        self.health = 100
        self.dead = False
        leg1, leg2, body, arm1, arm2 = self.sprites[0:5]
        body.setPosition((0, 5))
        leg1.setPosition((0, 10))
        leg2.setPosition((2, 10))
        arm1.setPosition((1, 7))
        arm2.setPosition((1, 7))
        self.climbHeight = 4
        self.xterm, self.yterm = 0.2, 0.3
        self.xacc, self.yacc = 0, 0
        self.xpacc, self.xnacc = 0, 0
        self.resting = False
        self.camera = TwoDVertexCamera()
        self.xpos, self.ypos, self.xvel, self.yvel = 20, 100, 0, 0
        self.width = 3
        self.height = 15
        self.jumping = False
        self.broadcastSystem = broadcastSystem
        for sprite in self.sprites:
            sprite.setCameras([self.camera, groundCamera])
            sprite.setFilter(0)
        self.broadcastSystem[K_UP].add(self.jump)
        self.broadcastSystem[K_LEFT].add(self.moveLeft)
        self.broadcastSystem[K_RIGHT].add(self.moveRight)
        self.scheduler = scheduler
        self.jumpCancel = ScheduledTask(self.cancelJump)
        
    def kill(self):
        """Kill the dude"""
        self.broadcastSystem[K_UP].remove(self.jump)
        self.broadcastSystem[K_LEFT].remove(self.moveLeft)
        self.broadcastSystem[K_RIGHT].remove(self.moveRight)
        del self.sprites
        self.dead = True
        
    def moveRight(self, keydown):
        """Move right handler"""
        self.xpacc = 0.2 if keydown else 0
        self.xacc = self.xnacc + self.xpacc
        
    def moveLeft(self, keydown):
        """Move left handler"""
        self.xnacc = -0.2 if keydown else 0
        self.xacc = self.xnacc + self.xpacc
        
    def cancelJump(self):
        """Stop jump acceleration handler"""
        if self.jumping:
            self.yacc = min(0, self.yacc)
            self.jumping = False
        
    def jump(self, keydown):
        """Jump handler"""
        if keydown:
            if self.resting:
                self.yacc = 1
                self.scheduler[140].add(self.jumpCancel)
                self.resting = False
                self.jumping = True
        else:
            self.cancelJump()
                    
        
class Ground(object):
    """The ground of the level"""
    def __init__(self, lvlheight, camera, fragCamera, imageStore):
        """Initialise and generate the ground layer"""
        self.sprites = []
        self.heights = defaultdict(lambda: 0)
        self.outlineHeights = defaultdict(lambda: 0)
        self.genHeights()
        self.genSprites(lvlheight, camera, fragCamera, imageStore)
    
    def genHeights(self):
        """Generate a dictionary containing the heights of positions for 
        this level"""
        land = ['flat', 'uphill', 'downhill']
        absence = ['hole']
        vert = ['wall', 'cliff']
        allf = land + absence + vert
        features = {
            'flat': (lambda x: (x, x), 100, allf),
            'uphill': (lambda x: (min(40, x + 1), min(40, x + 1)), 20, allf),
            'downhill': (lambda x: (max(1, x - 1), max(1, x - 1)), 20, allf),
            'hole': (lambda x: (0, x), 4, land),
            'wall': (lambda x: (min(40, x + 4), min(40, x + 4)), 1, land),
            'cliff': (lambda x: (max(1, x - 8), max(1, x - 8)), 1, land),
            }
                    
        currentFeature = 'flat'
        featureChangeProb = 0.1
        currentHeight = 20
        currentPos = 0
        featureFunc, featureLenRemaining, nextFeatures = \
            features[currentFeature]
        
        for _ in xrange(10):
            # generate flat at beginning
            self.heights[currentPos], currentHeight = featureFunc(currentHeight)
            currentPos += 1
            featureLenRemaining -= 1
            
        for _ in xrange(500):
            if random.random() < featureChangeProb or not featureLenRemaining:
                currentFeature = random.choice(nextFeatures)
                featureFunc, featureLenRemaining, nextFeatures = \
                    features[currentFeature]
                nextFeatures = nextFeatures[:]
                if currentFeature in nextFeatures:
                    nextFeatures.remove(currentFeature)
            self.heights[currentPos], currentHeight = featureFunc(currentHeight)
            featureLenRemaining -= 1 
            currentPos += 1
            
    def genSprites(self, lvlheight, camera, fragCamera, imageStore):
        """Generate the sprites for the hieghts in the dictionary"""
        for x in xrange(min(self.heights), max(self.heights)):
            height = self.heights[x]
            xcoord = 8 * x
            ycoord = lvlheight - (8 * height)
            if self.heights[x-1] < height and self.heights[x+1] < height:
                topimage = imageStore('tipblock')
                for z in xrange(4):
                    self.outlineHeights[xcoord + z] = 8 * (height - 1) + z
                for z in xrange(4, 8):
                    self.outlineHeights[xcoord + z] = 8 * height - z
            elif self.heights[x-1] < height:
                topimage = imageStore('slopeblock')
                for z in xrange(8):
                    self.outlineHeights[xcoord + z] = 8 * (height - 1) + z
            elif self.heights[x+1] < height:
                topimage = imageStore('slopeblock2')
                for z in xrange(8):
                    self.outlineHeights[xcoord + z] = 8 * height - z
            else:
                topimage = imageStore('topblock')
                for z in xrange(8):
                    self.outlineHeights[xcoord + z] = 8 * height
            baseimage = imageStore('baseblock')
            
            if height != 0:
                self.sprites.append(Sprite(position=(xcoord, ycoord), 
                    image=topimage, offsets=topLeftOffsets, offsetsRel=True,
                    layer=-1, cameras=[camera], fragCameras=[fragCamera]))
                
                times = int((lvlheight - ycoord) / 8)
                ycoord += 8
                self.sprites.append(Sprite(position=(xcoord, ycoord),
                        image=baseimage, offsets=topLeftOffsets, 
                        offsetsRel=True, tiling = (1, times),
                        layer=-1, cameras=[camera], fragCameras=[fragCamera]))
                            

class HealthDisplay(object):
    """Display the dudes (or others) health - currently not really used"""
    def __init__(self, track, zoom, broadcastSystem,imageStore):
        """Initialise the display"""
        self.tracked = track
        self.currentDisplay = self.tracked.health
        self.images = dict([(str(x), imageStore(str(x))) for x in xrange(10)])
        self.images['%'] = imageStore('percent')
        self.camera = TwoDVertexCamera()
        self.camera.setCameraScale(zoom)
        cameras = [self.camera]
        self.digit0 = Sprite(image=self.images['1'], position=(0, 0), 
                             layer=-1, cameras=cameras)
        self.digit1 = Sprite(image=self.images['0'], position=(9, 0), 
                             layer=-1, cameras=cameras)
        self.digit2 = Sprite(image=self.images['0'], position=(18, 0), 
                             layer=-1, cameras=cameras)
        self.percent = Sprite(image=self.images['%'], position=(27, 0), 
                              layer=-1, cameras=cameras)
        broadcastSystem['tick'].add(self.tick)
    
    def tick(self, dt):
        """Perform a tick"""
        if self.tracked.health != self.currentDisplay:
            self.currentDisplay = self.tracked.health
            strdisplay = str(self.currentDisplay).rjust(3, '0')
            for x, sprite in ((0, self.digit0), (1, self.digit1), 
                             (2, self.digit2)):
                sprite.setImage(self.images[strdisplay[x]])
        
    
class Baddie(object):
    """Unimplemented class for baddies"""
    
class Game(object):
    """Main object for the game"""
    def __init__(self):
        """Initialise the game"""
        self.windowSize = (800, 400)

    def main(self):
        """Main loop for the game"""
        GLCommon.initEnvironment(self.windowSize)
        broadcastSystem = Broadcaster()
        imageStore = ResourceStore(loadImage, 'data', '.png')
        SpriteCommon.setDefaultFilter('nearest')
        clock = pygame.time.Clock()
        groundEnemyCam = TwoDVertexCamera()
        fadingCam = TwoDFragmentCamera()
        zoom = 4.0
        scheduler = Scheduler(0, 5000, broadcastSystem)
        x = Ground(self.windowSize[1], groundEnemyCam, fadingCam, imageStore)
        d = Dude(groundEnemyCam, scheduler, broadcastSystem, imageStore)
        camManager = CameraManager(d, [groundEnemyCam], zoom,
                                   self.windowSize, broadcastSystem)
        health = HealthDisplay(d, zoom, broadcastSystem, imageStore)
        world = World(x, [d], broadcastSystem)
        going = True
        while going:
            t = clock.tick(60)
            broadcastSystem['tick'].broadcast(t)
            for event in pygame.event.get():
                if event.type == QUIT:
                    going = False
                elif event.type == KEYDOWN and event.key == K_ESCAPE:
                    going = False
                elif event.type in (KEYDOWN, KEYUP):
                    if event.key in broadcastSystem:
                        down = event.type == KEYDOWN
                        broadcastSystem[event.key].broadcast(down)
            pygl3display.GLCommon.doTasks()
            if d.dead:
                going = False
        sys.exit(0)

def main():
    """Run the game"""
    x = Game()
    x.main()
    
if __name__ == '__main__':
    if 'debug' in sys.argv:
        DEBUG = True
    if 'profile' in sys.argv:
        import cProfile
        cProfile.run('main()')
    else:
        main()

    
