"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.examples.textexample

This file used to provide an example of most of the high performance usage of
PyGL3Display 2d Sprites in version 1.1 and below. In Version 1.9 and onwards,
it exists to demonstrate SpriteGroups, even though these are under threat of
depreciation.
"""

from exampleskeleton import ExampleSkeleton, pyg3
import pygame
from pygame.locals import (KEYDOWN, K_r, K_g, K_b, K_DOWN, K_UP, K_LEFT, 
    K_RIGHT, K_y, K_i, K_j, K_k, K_l)
from pygl3display.sprites.pygameimages import PygameSurface
    
class SpriteFont(object):
    """Renders a font into a series of sprites, and handles how to
    get these printed onto screen"""
    def __init__(self, font, size):
        """Prerender the font"""
        chars  = ' abcdefghijklmnopqrstuvwxyzQWERTYUIOPASDFGHJKLZXCVBNM,.'
        chars += '/?!"$%^\\\';&*()-_=+1234567890'
        font = pygame.font.Font(pygame.font.match_font(font), size)
        self.fontmap = {}
        self.height = 0
        for char in chars:
            character = font.render(char, True, (255, 255, 255, 255))
            self.height = max((self.height, character.get_height()))
            self.fontmap[char] = (PygameSurface.surfaceify(character), 
                                 character.get_width())
    
    def makeString(self, string):
        """Creates a SpriteGroup which renders the string."""
        lines = string.split('\n')
        sprites = []
        ypos = 0
        maxXpos = 0
        for line in lines:
            xpos = 0
            for char in line:
                if char in self.fontmap:
                    if char != ' ':
                        newSprite = pyg3.Sprite(self.fontmap[char][0])
                        newSprite.setOffsets(pyg3.topLeftOffsets, relative=True)
                        newSprite.setPosition((xpos, ypos))
                        sprites.append(newSprite)
                    xpos += self.fontmap[char][1]
            maxXpos = max(xpos, maxXpos)
            ypos += self.height
        ret = pyg3.SpriteGroup(sprites)
        return ret

class SpriteFontApp(ExampleSkeleton):
    """An example demonstrating Sprites and SpriteGroups"""
    
    def __init__(self):
        """Initialise the example"""
        super(SpriteFontApp, self).__init__('Text example', (800, 800))
        self.string = ('Things you can do: \n' 
                       ' r: cycle red values \n' 
                       ' g: cycle green values \n' 
                       ' b: cycle blue values \n' 
                       ' arrow keys: move text \n' 
                       ' y: clockwise rotate \n' 
                       ' ijkl: increase / decrease x/y scale\n' 
                       'I\'m afraid that holding keys down won\'t work.\n' 
                       'Apologies for the problems with this sample.\n ' 
                       '      Habilain')
        self.origx, self.origy = (0, 0)
        self.red, self.green, self.blue = (255, 255, 255)
        self.posx, self.posy = (0, 0)
        self.scalex, self.scaley = (1, 1)
        self.rotation = 0
        self.sinceLastCount = 0
        self.frames = 0
        self.countedTime = 0
        
    def init(self):
        """Initialise the font and put text on screen"""
        self.font = SpriteFont('dejavu-sans', 20)
        self.text = self.font.makeString(self.string)
        
    def eventHandler(self, event):
        """An event handler to move the text around"""
        if event.type == KEYDOWN:
            if event.key == K_r:
                self.red = (self.red + 10) % 256
                self.text.setColor((self.red, self.blue, self.green, 255))
            elif event.key == K_g:
                self.green = (self.green + 10) % 256
                self.text.setColor((self.red, self.blue, self.green, 255))
            elif event.key == K_b:
                self.blue = (self.blue + 10) % 256
                self.text.setColor((self.red, self.blue, self.green, 255))
            elif event.key == K_DOWN:
                self.posy += 10
                self.text.setPosition((self.posx, self.posy))
            elif event.key == K_UP:
                self.posy -= 10
                self.text.setPosition((self.posx, self.posy))
            elif event.key == K_RIGHT:
                self.posx += 10
                self.text.setPosition((self.posx, self.posy))
            elif event.key == K_LEFT:
                self.posx -= 10
                self.text.setPosition((self.posx, self.posy))
            elif event.key == K_y:
                self.rotation += 0.1
                self.text.setRotation(self.rotation)
            elif event.key == K_i:
                self.scaley -= 0.1
                self.text.setScaleY(self.scaley)
            elif event.key == K_k:  
                self.scaley += 0.1
                self.text.setScaleY(self.scaley)
            elif event.key == K_j:
                self.scalex -= 0.1
                self.text.setScaleX(self.scalex)
            elif event.key == K_l:
                self.scalex += 0.1
                self.text.setScaleX(self.scalex)
                
    def loop(self, dt):
        """Take logging and update fps in title bar"""
        self.sinceLastCount += dt
        if self.sinceLastCount > 500:
            self.sinceLastCount -= 500
            self.countedTime += 0.5
            caption = 'torture test pygame (instant) fps ' + \
                      str(round(self.clock.get_fps(), 2)) + ' averaged fps '\
                      + str(round((self.frames / self.countedTime), 2)) 
            pygame.display.set_caption(caption)
        self.frames += 1

        
if __name__ == '__main__':
    SpriteFontApp().run()
