"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.fbo

Provides access to FBOs.
"""

from collections import defaultdict

from pygl3display.glcommoninst import GLCommon
from pygl3display.gl import (glGetInteger, glDrawBuffers, 
    glCheckFramebufferStatus, glBindFramebuffer, glFramebufferTexture2D, 
    GL_FRAMEBUFFER, GL_FRAMEBUFFER_COMPLETE, GL_MAX_COLOR_ATTACHMENTS, 
    GL_TEXTURE_2D, GL_NONE)
from pygl3display.gl import (ColorAttachmentDict as ColorAttachmentDict, 
                             FBOENABLED as ENABLED)
from pygl3display.genstore import Framebuffers

class FBO(object):
    """An abstraction of an FBO"""
    def __init__(self):
        """Initialise the FBO"""
        self.framebuffer = Framebuffers.getOne(self) #glGenFramebuffer()
        
    def checkErrors(self):
        """Check to see if the FBO has any errors; if so, return error code,
        otherwise return True"""
        FBOCommonInst.bindFramebuffer(self)
        fbostatus = glCheckFramebufferStatus(GL_FRAMEBUFFER)
        return True if fbostatus == GL_FRAMEBUFFER_COMPLETE else fbostatus
        
    def bind(self):
        """Bind the framebuffer"""
        FBOCommonInst.bindFramebuffer(self)
    
    @staticmethod
    def unbind():
        """Unbinds framebuffers"""
        FBOCommonInst.bindFramebuffer(None)
    
    def bindTextures(self, textures):
        """Bind a given list of textures"""
        FBOCommonInst.bindFramebuffer(self)
        buffers = []
        for texture, indx in zip(textures, xrange(len(textures))):
            location = FBOCommonInst.bindPoints.get(u'color%i' % indx)
            glFramebufferTexture2D(GL_FRAMEBUFFER, location,
                GL_TEXTURE_2D, texture.getTexture(), 0)
            buffers.append(location)
        glDrawBuffers(buffers)
        
    def unbindTextures(self, number):
        """Unbind a number of textures"""
        FBOCommonInst.bindFramebuffer(self)
        for indx in xrange(number):
            location = FBOCommonInst.bindPoints.get(u'color%i' % indx)
            glFramebufferTexture2D(GL_FRAMEBUFFER, location,
                GL_TEXTURE_2D, 0, 0)
        glDrawBuffers([])
        
if ENABLED:
    class FBOCommon(object):
        """Shared state between FBO objects"""
        def __init__(self):
            """Initialise the shared state"""
            self.boundFBO = 0
            self.fbos = defaultdict(FBO)
            self.boundBindPoints = []

        def __getattr__(self, key):
            """Enable lazy creation of the bindPoints dictionary"""
            if key in (u'bindPoints', u'noColorBindPoints'):
                self.getBindPoints()
                return self.bindPoints
            else:
                return super(FBOCommon, self).__getattribute__(key)

        def getFBO(self, size):
            """Get the shared FBO for a given size; all textures in an
            FBO need to be the same size for FBOs to work on the OpenGL
            extension (as opposed to the OpenGL 3 core)"""
            return self.fbos[size]
                            
        def getBindPoints(self):
            """Construct the locations dictionary"""
            self.noColorBindPoints = glGetInteger(GL_MAX_COLOR_ATTACHMENTS)
            self.bindPoints = {}
            colorAttachmentDict = ColorAttachmentDict()
            for x in xrange(self.noColorBindPoints):
                self.bindPoints[u'color%i' % x] = \
                    colorAttachmentDict[u'GL_COLOR_ATTACHMENT%i' % x]
                
        def bindFramebuffer(self, fbo):
            """Lazily bind a framebuffer"""
            GLCommon.target = fbo
            fboval = GL_NONE if fbo is None else fbo.framebuffer
            if self.boundFBO != fboval:
                glBindFramebuffer(GL_FRAMEBUFFER, fboval)
                self.boundFBO = fbo
    
    FBOCommonInst = FBOCommon()
    
    def unbindFBO():
        """Unbind any bound FBO; a hook for GLCommon, but needs to be in the
        FBO file as OpenGL 2.1 does not guarantee FBO's available"""
        FBOCommonInst.bindFramebuffer(None)
    GLCommon.addSwitchToMainTargetHook(unbindFBO)

