"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3).

pygl3display.genstore

A weak store for GL ID's. These are given out to a user and remain with the
user object until the user is cleared.
"""

import weakref
from pygl3display.gl import (glGenBuffers, glGenFramebuffers)

class GenStore(object):
    """A manager for the IDs handed out by glGen functions"""
    def __init__(self, genFunc):
        """Set up the manager for the given function"""
        self.genFunc = genFunc
        self.free = set()
        self.bound = {}
        
    def getOne(self, user):
        """Get a single id
        Args:
          user: the object that will use the id"""
        return self.get(user, 1)[0]
    
    def get(self, user, n):
        """Get a list of ids
        Args:
          user: the object that will use the id
          n: the number of ids to get"""
        ret = []
        while self.free and n > 0:
            ret.append(self.free.pop())
            n -= 1
        if n > 0:
            ret.extend(self.genFunc(n))
        userref = weakref.ref(user, self.__release)
        self.bound[userref] = ret
        return ret
    
    def __release(self, userref):
        """Internally used to release ids once the object
        using them expires"""
        self.free.update(self.bound[userref])
        del self.bound[userref]
        
Buffers = GenStore(glGenBuffers)
Framebuffers = GenStore(glGenFramebuffers)