"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance Pygame OpenGL library, capable of handling 1000s of sprites

Requirements: Python, Pygame, PyOpenGL, Open GL 2.0 graphics hardware

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). By using this file you agree
to be bound by at least one of these licenses.

Other licenses, including commercial/proprietary licenses, can be arranged by 
contacting me via e-mail at habilain@gmail.com. If you do not have in physical 
writing an alternative license agreement from me, you must use one of the 
licenses specified above when releasing work which makes use of this library.

pygl3display.gl.cffigl.cffigl_dev_share

Python port of http://www.opengl-tutorial.org/ LoadShaders function"""
from __future__ import division
import math, itertools, operator, png

def readPNG(fn):
    reader = png.Reader(filename=fn)
    width, height, pixels = reader.read_flat()[:3]
    return width, height, list(pixels)

def LoadShaders(gl, vertexShader, fragmentShader):
    vertexShaderID = gl.glCreateShader(gl.GL_VERTEX_SHADER)
    fragmentShaderID = gl.glCreateShader(gl.GL_FRAGMENT_SHADER)
    gl.glShaderSource(vertexShaderID, [vertexShader])
    gl.glCompileShader(vertexShaderID)
    result = gl.glGetShaderi(vertexShaderID, gl.GL_COMPILE_STATUS)
    if result != gl.GL_TRUE:
        print gl.glGetShaderInfoLog(vertexShaderID)
        raise Exception
    gl.glShaderSource(fragmentShaderID, [fragmentShader])
    gl.glCompileShader(fragmentShaderID)
    result = gl.glGetShaderi(fragmentShaderID, gl.GL_COMPILE_STATUS)
    if result != gl.GL_TRUE:
        print gl.glGetShaderInfoLog(vertexShaderID)
        raise Exception
    programID = gl.glCreateProgram()
    gl.glAttachShader(programID, vertexShaderID)
    gl.glAttachShader(programID, fragmentShaderID)
    gl.glLinkProgram(programID)
    if gl.glGetProgrami(programID, gl.GL_LINK_STATUS) != gl.GL_TRUE:
        print gl.glGetProgramInfoLog(programID)
        raise Exception
    gl.glDeleteShader(vertexShaderID)
    gl.glDeleteShader(fragmentShaderID)
    return programID

class Matrix(object):
    def __init__(self, rows, columns, content=None):
        self.rows = rows
        self.columns = columns
        if content is not None:
            if len(content) != rows or any((len(x) != columns for x in content)):
                raise Exception('Content doesn\'t match description')
            else:
                self.content = content 
        else:
            self.content = [[0.0]*columns for _ in xrange(rows)]
            
    def data(self):
        """Returns list of elements suitable for OpenGL"""
        r =  list(itertools.chain(*self.content))
        return r
        
    def __str__(self):
        maxStrLen = 0
        for x in self.content:
            for y in x:
                maxStrLen = max(maxStrLen, len(str(y)))
            
        return '\n'.join(['[%s]' % ', '.join([str(y).rjust(maxStrLen) for y in x]) for x in self.content])
    
    def __multypeop__(self, other, op):
        if isinstance(other, Matrix) and self.columns == other.rows:
            ret = Matrix(self.rows, other.columns)
            scontent, ocontent = self.content, other.content
            rcontent = ret.content
            for x in xrange(self.rows):
                for y in xrange(other.columns):
                    values = [op(scontent[x][z], ocontent[z][y]) for z 
                              in xrange(self.rows)]
                    rcontent[x][y] = sum(values)
            return ret
        elif isinstance(other, (int, float)):
            return Matrix(self.rows, self.columns, 
                              [[op(x, other) for x in y] for y in self.content])
        else:
            raise Exception('Cannot multiply these matricies!')

    def __addtypeop__(self, other, op):
        if isinstance(other, Matrix) and self.columns == other.columns and \
        self.rows == other.rows:
            return Matrix(self.rows, self.columns,
                    [[op(self.content[x][y], other.content[x][y]) 
                      for y in xrange(self.columns)] for x in xrange(self.rows)])
        elif isinstance(other, (int, float)):
            return Matrix(self.rows, self.columns, 
                              [[op(x, other) for x in y] for y in self.content])

    def __mul__(self, other):
        return self.__multypeop__(other, operator.mul)
    
    def __rmul__(self, other):
        return other.__multypeop__(self, operator.mul)
    
    def __div__(self, other):
        return self.__multypeop__(other, operator.div)
    
    def __rdiv__(self, other):
        return other.__multypeop__(self, operator.div)
        
    def __add__(self, other):
        return self.__addtypeop__(other, operator.add)
    
    def __radd__(self, other):
        return other.__addtypeop__(self, operator.add)
                    
    def __sub__(self, other):
        return self.__addtypeop__(other, operator.sub)
    
    def __rsub__(self, other):
        return other.__addtypeop__(self, operator.sub)
            
    def normalise(self):
        mx, mn = float('-inf'), float('inf')
        for x in self.content:
            mx = max(mx, max(x))
            mn = min(mn, min(x))
        rnge = mx - mn
        if rnge != 0:
            return Matrix(self.rows, self.columns, 
                          [[(y - mn) / rnge for y in x] for x in self.content])
        else:
            # Normalise to 0
            return Matrix(self.rows, self.columns)
        
    def transpose(self):
        rets = {}
        for x in self.content: 
            for y in xrange(len(x)):
                if y not in rets: rets[y] = []
                rets[y].append(x[y])
        m = []
        for x in sorted(rets.keys()):
            m.append(rets[x])
        return Matrix(self.columns, self.rows, m)
            
class Identity(Matrix):
    def __init__(self, size):
        super(Identity, self).__init__(size, size, 
                [[1 if x == y else 0 for y in xrange(size)] for x in xrange(size)])

class Vector(Matrix):
    def __init__(self, rows, content=None):
        super(Vector, self).__init__(rows, 1, 
                    [[x] for x in content] if content is not None else None)
        
            
def cross(vec1, vec2):
    if (isinstance(vec1, Matrix) and isinstance(vec2, Matrix) and 
        vec1.rows == 3 and vec2.rows == 3 and vec1.columns == 1 and vec2.columns == 1):
        v1c = vec1.transpose().content[0]
        v2c = vec2.transpose().content[0]
        return Vector(3, [v1c[1]*v2c[2] - v1c[2]*v2c[1], 
                          v1c[2]*v2c[0] - v1c[0]*v2c[2],
                          v1c[0]*v2c[1] - v1c[1]*v2c[0],])
    else:
        raise Exception('Require vectors of length 3 to cross product')
    
def dot(vec1, vec2):
    if (isinstance(vec1, Matrix) and isinstance(vec2, Matrix) and 
        vec1.rows == vec2.rows == 3):
        v1c = vec1.transpose().content[0]
        v2c = vec2.transpose().content[0]
        return sum((x*y for x,y in zip(v1c, v2c)))
    else:
        raise Exception('Require vectors to dot product')
            
def normalise(vec):
    if isinstance(vec, Matrix) and vec.columns == 1:
        sqr = dot(vec, vec)
        return vec * (1.0 / math.sqrt(sqr))
    
def perspective(fov, aspect, znear, zfar):
    """fov in radians, aspect, znear, zfar as per normal"""
    f = 1.0 / math.tan(fov / 2.0)
    q = (zfar + znear) / (znear - zfar)
    q2 = (2 * zfar * znear) / (znear - zfar)
    m = [[f / aspect, 0,  0,  0],
         [         0, f,  0,  0],
         [         0, 0,  q, -1],
         [         0, 0, q2,  0]]
    return Matrix(4, 4, m)

def lookat(eye, center, up):
    f = normalise(center - eye)
    u = normalise(up)
    s = normalise(cross(f, u))
    u = cross(s, f)
    sc, uc, fc = s.transpose().content[0], u.transpose().content[0], f.transpose().content[0]
    result = Matrix(4, 4, 
                    [[sc[0], uc[0], -fc[0], 0.0],
                     [sc[1], uc[1], -fc[1], 0.0],
                     [sc[2], uc[2], -fc[2], 0.0],
                     [-dot(s, eye), -dot(u, eye), dot(f, eye), 1.0]])
    return result

def vec3(*args): return Vector(3, args)

def makeWindow(bindings, size, title):
    gl = bindings
    if not gl.glfwInit():
        raise Exception('Could not initialise GLFW')
    gl.glfwOpenWindowHint(gl.GLFW_FSAA_SAMPLES, 4)
    gl.glfwOpenWindowHint(gl.GLFW_OPENGL_VERSION_MAJOR, 2)
    gl.glfwOpenWindowHint(gl.GLFW_OPENGL_VERSION_MINOR, 1)
    if not gl.glfwOpenWindow(size[0], size[1], 0,0,0,0, 32,0, gl.GLFW_WINDOW ):
        gl.glfwTerminate()
        raise Exception('Could not create GL window')
    glewret = gl.glewInit()
    if glewret != gl.GLEW_OK:
        gl.glfwTerminate()
        raise Exception('Could not initialise GLEW')
    gl.glfwSetWindowTitle(title)

    