"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.gl3shader

This file provides the basic shader class that all PyGL3Display shaders
should derive from.
"""

from __future__ import division
import weakref
from collections import defaultdict
from pygl3display.glcommon import ShaderError
from pygl3display.glcommoninst import GLCommon
from pygl3display.gl import (
     glCreateProgram, glLinkProgram, glAttachShader,  
     glGetAttribLocation, glGetUniformLocation, 
     glDeleteProgram, glDeleteShader, glGetProgrami, glGetProgramInfoLog, 
     GL_VERTEX_SHADER, GL_FRAGMENT_SHADER, GL_LINK_STATUS, GL_FALSE)
from pygl3display.noneweakkey import NoneWeakKeyDict

# TODO: There will likely have to be a bit of work done to get this
# working with Unicodes.

class GL3Shader(object):
    """Abstraction of a shader program and the vertex arrays which 
    support it"""
    
    def __init__(self, vertexProgs, fragmentProgs, **kwargs):
        """Initialise the shader.
        Arguments:
         vertexProgs  : A list of vertex programs (source code)
         fragmentProgs: A list of fragment programs (source code)"""
        self.vertexProgs = vertexProgs
        self.fragmentProgs = fragmentProgs
        fullProg = '\n'.join(vertexProgs + fragmentProgs)
        arrays = set()
        uniforms = set()
        for line in fullProg.split('\n'):
            line = line.strip()
            if line.startswith('attribute'):
                arrays.add(line.partition(' ')[2].partition(' ')[2]
                    .partition(';')[0].partition('[')[0].strip())
            elif line.startswith('uniform'):
                uniforms.add(line.partition(' ')[2].partition(' ')[2]
                    .partition(';')[0].partition('[')[0].strip())
        self.arrayNames = arrays
        self.uniforms = uniforms
        self.setup()
        self.clientLayers = defaultdict(NoneWeakKeyDict)
        self.clients = weakref.WeakKeyDictionary()
        GLCommon.addShader(self)
        
    def addClient(self, client):
        """Add a client to the shader"""
        self.clients[client] = True
        
    def addClientLayer(self, client, layer, target):
        """Allow a client to draw on a layer
        Arguments:
         client : The client to allow
         layer : The layer to allow to draw on
         target: The target to allow to draw on"""
        if (layer not in self.clientLayers 
           or target not in self.clientLayers[layer]): 
            self.clientLayers[layer][target] = weakref.WeakKeyDictionary()
        self.clientLayers[layer][target][client] = True

    def removeClientLayer(self, client, layer, target):
        """Disallow a client to draw on a layer; this is for performance,
        if it gets used.
        Arguments:
         client: The client to disallow
         layer: The layer to disallow drawing onto
         target: The target to allow to draw on"""
        if (layer in self.clientLayers and 
        target in self.clientLayers[layer] and 
        client in self.clientLayers[layer][target]): 
            del self.clientLayers[layer][target][client]

    def setup(self):
        """Perform the setup of the actual shader program, and the necessary
        setup of data for binding arrays/uniforms."""
        
        self._vertexShaders = [
           GLCommon.getShader(vertexProg, GL_VERTEX_SHADER) for 
           vertexProg in self.vertexProgs]
        self._fragmentShaders = [
           GLCommon.getShader(fragmentProg, GL_FRAGMENT_SHADER) for 
           fragmentProg in self.fragmentProgs]
                                 
        self.shaderProgram = glCreateProgram()
        for vertexShader in self._vertexShaders:
            glAttachShader(self.shaderProgram, vertexShader)
        for fragmentShader in self._fragmentShaders:
            glAttachShader(self.shaderProgram, fragmentShader)
        glLinkProgram(self.shaderProgram)
        if glGetProgrami(self.shaderProgram, GL_LINK_STATUS) == GL_FALSE:
            infoLog = glGetProgramInfoLog(self.shaderProgram)
            errorMessage = 'Shader Link Error:\n Error Message:' + infoLog
            raise ShaderError(errorMessage)
        
        GLCommon.useProgram(self)
        
        self.arrayLocs = dict([(key, 
            glGetAttribLocation(self.shaderProgram, key)) 
            for key in self.arrayNames])
        self.uniformLocs = dict([(key, 
                             glGetUniformLocation(self.shaderProgram, key)
                            ) for key in self.uniforms])

    def recreate(self):
        """Reset the shader, and all managers attached to it"""
        glDeleteProgram(self.shaderProgram)
        for shader in self._vertexShaders:
            glDeleteShader(shader)
        for shader in self._fragmentShaders:
            glDeleteShader(shader)
        self.setup()
        for ref in self.clients.keyrefs():
            if ref(): 
                ref().recreate()      
            
    def doTasks(self, layer, target):
        """Perform all tasks on a given layer
        Arguments:
         layer : the layer to draw
         target: the target to draw"""
        if layer in self.clientLayers and target in self.clientLayers[layer]:
            if GLCommon.boundShader is not self:
                GLCommon.useProgram(self)
            for ref in self.clientLayers[layer][target].keyrefs():
                if ref():
                    ref().doTasks(layer, target)
                
