"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.infrastructure

This file provides the basic infrastructure of the PyGL3Display project.
"""

from __future__ import division
from collections import defaultdict
from pygl3display.glcommoninst import GLCommon
from pygl3display.slices import Slices
from pygl3display.gl import glDrawElements, byref
from pygl3display.noneweakkey import NoneWeakKeyDict
import weakref


class GL3Layer(object):
    """GL3Layer represents a 'layer' of the rendering, i.e. one pass of
    a shader."""
    def __init__(self, layer, vertexArrays):
        """Initialise the layer.
        Arguments:
            layer: The name of the layer
            vertexArrays: The VertexArrays object that the layer is stored on 
        """
        self.indicesLength = int(vertexArrays.noSlots * 
                                 vertexArrays.slotMultiplier)
        self.indicesGen = lambda: vertexArrays.indicesCType(self.indicesLength)
        self.sizeOfIndices = vertexArrays.sizeOfIndices
        self.indicesData = []
        GLCommon.updateLayers(layer)
        
    def __iter__(self):
        """Enable iteration of the layer"""
        return iter(self.indicesData)
    
    def __getIndices(self, n):
        """Return a GL3Slots of n slots. If they cannot be allocated
        on one the currently existing sets of indices, a new set
        is created."""
        for indices in self.indicesData:
            if indices.canAllocate(n):
                return indices
        newIndices = GL3Indicies(self)
        if not newIndices.canAllocate(n):
            raise Exception(u'Cannot allocate %s indicies' % str(n))
        self.indicesData.append(newIndices)
        return newIndices
    
    def getRenderSlots(self, n):
        """Sets up and returns a GL3Slots of n render slots"""
        return self.__getIndices(n).allocate(n)
    
    def getRenderSlotBlocks(self, n, blocks):
        """Return a list of contiguous GL3Slots objects, each containing
        n slots, blocks elements long."""
        return self.__getIndices(n * blocks).allocateBlocks(n, blocks)
        
class GL3Slots(object):
    """Holds slots allocated by GL3Indicies; also used in providing
    a hook for a weakref based deallocation."""
    def __init__(self, slots, indices):
        """Initialise the GL3Slots
        Args:
          slots: The slots this slots object should hold
          indices: The GL3Indices object that allocated the slots"""
        self.slots = tuple(slots)
        self.indices = indices
        
    def __str__(self):
        """Debug printing method"""
        return 'GL3Slots %s on %s containing %s' % (str(self.slots), 
                str(self.indices), str(self.indices.content(self.slots)))
        
    def release(self):
        """Release the slots contained back to the indices store"""
        self.indices.release(self.slots)
        
    def upload(self, data):
        """Upload data onto the render slots"""
        self.indices.upload(self.slots, data)
    
class GL3Indicies(object):
    """Stores a set of indices used for the index array"""
    def __init__(self, layerData):
        """Makes a list to store one set of indices on the given layer.
        Arguments:
          layerData: The layer data to be bound to
        """
        self.indices = layerData.indicesGen()
        self.len = 0
        self.slots = Slices(0, layerData.indicesLength)
        self.allocated = set()
        self.dirty = False
        self.pending = []
        self.startAddress = byref(self.indices, 0)
        self.refs = {}
        
    def canAllocate(self, n):
        """Query if it is possible to allocate n render slots"""
        return self.slots.canAllocate(n)
    
    def __allocate(self, n):
        """Get a list of n slots available"""
        self.dirty = True
        newAlloc = self.slots.allocate(n)
        self.allocated.update(newAlloc)
        return newAlloc
    
    def __setTeardown(self, glslots):
        """Set the teardown procedure on glslots"""
        ref = weakref.ref(glslots, self.release)
        self.refs[ref] = glslots.slots
        return glslots
    
    def release(self, ref):
        #TODO: Should this be __release? Doesn't seem to be used elsewhere
        """Release slots associated with a given weakref, used by
        __setTearDown"""
        if ref in self.refs:
            slots = self.refs[ref]
            self.dirty = True
            self.pending.extend(slots)
            self.slots.release(slots)
            del self.refs[ref]
        
    def allocate(self, n):
        """Allocate and set a teardown on n slots; returns a GL3Slots"""
        td = self.__setTeardown
        return td(GL3Slots(self.__allocate(n), self))
    
    def allocateBlocks(self, n, blocks):
        """Allocate and set teardowns on n*blocks contiguous slots,
        split into blocks GL3Slots. Returns a list of GL3Slots"""
        td = self.__setTeardown
        slots = self.__allocate(n * blocks)
        divSlots = [slots[6*x:6*x+6] for x in xrange(int(len(slots)/6))]
        return [td(GL3Slots(x, self)) for x in divSlots]

    def clean(self):
        """Perform housekeeping;
        doTasks should call this if the attribute dirty is True"""
        for slot in self.pending: 
            self.indices[slot] = 0
        self.slots.join()
        self.allocated.difference_update(self.pending)
        self.pending = []
        self.dirty = False
        if self.allocated:
            minimum = min(self.allocated)
            maximum = max(self.allocated)
            self.len = maximum - minimum + 1
            self.startAddress = byref(self.indices, minimum)
        else:
            self.len = 0
            
    def content(self, slots):
        """Return the content of the index array, for debug purposes"""
        return [self.indices[x] for x in slots]
        
    def upload(self, slots, data):
        """Upload data onto given slots"""
        for x in xrange(len(slots)):
            self.indices[slots[x]] = data[x]
            
class GL3Client(object):
    """GL3Client is a base class for items which interact directly with the 
    manager. The clients track all GL state, and handle the actual drawing 
    of primitives to screen."""
    
    def __init__(self, glType, shader, vertexArraySpec, uniforms):
        """Initialise the GL3Client.
        Args:
         glType : The GL Type of the data to be rendered (e.g. GL_LINES)
         shader : the shader which created this client
         vertexArraySpec : An array specification for the vertex arrays this
                           client will be drawing"""
        self.vertexArraySpec = vertexArraySpec
        self.vertexArraysD = {} 
        self.shader = shader
        self.layers = defaultdict(lambda: defaultdict(NoneWeakKeyDict))
        self.renderHooks = []
        self.glType = glType
        shader.addClient(self)
        self.uniforms = uniforms
        
    def addRenderHook(self, callback):
        """Add a function to be called from the client when rendering."""
        self.renderHooks.append(callback)
            
    def doTasks(self, layer, target):
        """Instuct the client to draw everything it is responsible for to
        screen, for the given layer. 
        Will also perform housekeeping as necessary."""
        glType = self.glType
        for vertexArrays in self.layers:
            layerDict = self.layers[vertexArrays]
            if layer not in layerDict or target not in layerDict[layer]:
                pass
            else:
                for func in self.renderHooks:
                    func(layer)
                layerData = layerDict[layer][target]
                for indices in layerData:
                    if indices.dirty is True: 
                        indices.clean()
                    if indices.len == 0: 
                        pass
                    else:
                        self.uniforms.bind()
                        vertexArrays.bind()
                        glDrawElements(glType, indices.len, 
                            vertexArrays.indicesType, indices.startAddress)
        
    def getVertexArrays(self, n):
        """Return a vertex array that has at least n slots available"""
        ret = None
        for vertexArrays in self.vertexArraysD:
            if vertexArrays.canAllocDataSlots(n):
                ret = vertexArrays
                break
        if ret is None:
            ret = self.vertexArraySpec.makeArrays(65535)
            self.vertexArraysD[ret] = []
        return ret
        
    def __getLayerForRenderSlots(self, layer, target, vertexArrays):
        """Returns a correctly set up layerdata object for given layer, 
        target & vertexArrays."""
        if target not in self.layers[vertexArrays][layer]:
            layerData = GL3Layer(layer, vertexArrays)
            self.layers[vertexArrays][layer][target] = layerData
            self.shader.addClientLayer(self, layer, target)
        else:
            layerData = self.layers[vertexArrays][layer][target]
        return layerData
    
    def getRenderSlots(self, n, layer, target, vertexArrays):
        """Get n slots in the render array
        Always returns adjacent items
        Arguments:
         image: The image to allocate slots to
         n: The number of slots needed
         layer: The layer the slots should be on
        """
        layer = self.__getLayerForRenderSlots(layer, target, vertexArrays)
        return layer.getRenderSlots(n)
    
    def getRenderSlotBlocks(self, n, blocks, layer, target, vertexArrays):
        """Get n*blocks slots in the render array, split into multiple blocks
        Always returns adjacent items
        Arguments:
         image: The image to allocate slots to
         n: The number of slots per item needed
         blocks: The number of items needed
         layer: The layer the slots should be on
        """
        layer = self.__getLayerForRenderSlots(layer, target, vertexArrays)
        return layer.getRenderSlotBlocks(n, blocks)
                
    def recreate(self):
        """A method to be overridden for clients which manage specific
        GL state (e.g. texture atlases managing textures)"""
    