"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3).

pygl3display.noneweakkey

Provides a weak key dictionary extended so that it can also store a None value
"""

import weakref

class NoneWeakKeyDict(weakref.WeakKeyDictionary):
    """A minor modification of a weak key dictioanry that supports
    storing the none value as a key. This is necessary as "None" is a
    sensible value for having no render target, but WeakRefs are necessary
    to avoid memory leaks with actual render targets"""
    
    def __contains__(self, key):
        """Check if we have a None attribute"""
        if key is None:
            return hasattr(self, u'noneval')
        else:
            return weakref.WeakKeyDictionary.__contains__(self, key)
            
    def __getitem__(self, key):
        """Return the None value if necessary (note: exception raised for None
        not in dictionary is attribute error rather than key error - this is
        OK in PyGL3Display but poor practice in general)"""
        if key is None:
            return self.noneval
        else:
            return weakref.WeakKeyDictionary.__getitem__(self, key)
            
    def __setitem__(self, key, val):
        """Allow setting of a none key"""
        if key is None:
            self.noneval = val
        else:
            weakref.WeakKeyDictionary.__setitem__(self, key, val)
