"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.rendertargets

Provides TextureTarget and TwoDTextureTarget, which allow 
render to texture support.
"""

from pygl3display.texture.twod import TwoDTexture
from pygl3display.glcommoninst import GLCommon
from pygl3display.fbo import ENABLED

if not ENABLED:
    raise ImportError(u'FBOs not available')
    
from pygl3display.fbo import FBOCommonInst


class TextureTarget(object):
    """A Texture Target for render to multiple textures"""        
    
    def __init__(self, number, size, compute = False):
        """Initialise a texture render target
        Args:
         number: Number of output textures
         size: Size of textures
         compute: If true, use GL_RGBA32 textures and disable blending
         for the target."""
        self.size = size
        self.compute = compute
        self.fbo = FBOCommonInst.getFBO(size)
        self.compute = compute
        self.number = number
        self.initTextures()
                      
    def initTextures(self):
        """Initialise the textures of the texture target"""
        self.textures = [TwoDTexture(self.size, self.compute) 
                         for _ in xrange(self.number)]
        self.fbo.bindTextures(self.textures)
        status = self.fbo.checkErrors()
        if status is not True:
            raise Exception(u'FBO Status %i; not usable' % status)
        
    def getTextures(self, separate=False):
        """If separate is False, then the textures are replaced on the 
        target. If not, the textures persist on the render target, 
        and will be overwritten"""
        ret = self.textures[:]
        if separate:
            self.initTextures()
        return ret

    def __enter__(self):
        """Ensure that the render target is bound"""
        self.bind()
        
    def __exit__(self, val, etype, traceback):
        """Remove the bindings of the render target"""
        self.unbind()
        
    def bind(self):
        """Bind the fbo and the texture set this render target uses"""
        self.fbo.bind()
        self.fbo.bindTextures(self.textures)
        GLCommon.target = self
        GLCommon.setResolution((self.size, self.size))
        if self.compute:
            GLCommon.enableBlend(False)
        
    def unbind(self):
        """Unbind the texture set this render target uses"""
        self.fbo.unbindTextures(self.number)

        