"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.sprites.groups

This file provides the old, software driven, sprite group class. 
"""

from pygl3display.sprites.sprite import Sprite

class SpriteGroupMetaclass(type): 
    """Metaclass for SpriteGroups - but could be one day extracted for any
    grouping, because it deals with arbitary types"""
    
    def wrapFunc(cls, containedtype, funcname):
        """Assuming that the class is an iterable containing only
        containedtypes, generate a moderately high performance wrapper that
        applies the named function to everything in the iterable."""
        if hasattr(containedtype, funcname):
            func = getattr(containedtype, funcname)
            if callable(func):
                def wrapper(self, *args, **kwargs):
                    """Call the function on each item in the group"""
                    for sprite in self:
                        func(sprite, *args, **kwargs)
                setattr(cls, funcname, wrapper)

class SpriteGroup(object):
    """An object to group together Sprites, like a Pygame group.
    SpriteGroup present a similar API to a Sprite, with an operation
    applied to the group being applied to all contained sprites.
    
    Due to metaclass magic, all operators on a Sprite will work on a
    SpriteGroup. However, for the moment, some operations will produces odd
    results, most notably setOrigin.
    
    Note that when sprites are in a sprite group, they must share all of
    their attributes. Hence you should position sprites before adding them
    to the group, and then not touch them individually ever again.
    """
    __metaclass__ = SpriteGroupMetaclass
    
    def __init__(self, sprites=None):
        """Create a SpriteGroup
           Arguments:
                sprites: The sprite or sprites to be in the group initially
        """
        self._sprites = set()
        
        if sprites is not None:
            for container in sprites: 
                self.add(container)
        
    def __iter__(self):
        """Make the SpriteGroup an iterable"""
        return self._sprites.__iter__()
        
    def __len__(self):
        """Provide a way of determining the number of elements in the group"""
        return self._sprites.__len__()
        
    def __getattr__(self, key):
        """Try to wrap unknown function calls - any function defined on
        a Sprite will also work on a SpriteGroup, but some operations may
        have odd results (e.g. setOrigin)"""
        type(self).wrapFunc(Sprite, key)
        if hasattr(self, key):
            return getattr(self, key)
        else:
            return super(SpriteGroup, self).__getattribute__(key)
            
    def addSprite(self, sprite):
        """Add a single sprite to the Group"""
        if sprite not in self._sprites:
            self._sprites.add(sprite)
            spritePosX, spritePosY = sprite._position
            sprite._baseOffsets = [(offsetX+spritePosX, offsetY+spritePosY) 
                            for offsetX, offsetY in sprite._baseOffsets]
            sprite._position = (0, 0)
            sprite._internal.vertexArrays.setPosition(sprite._internal.slots, 
                                             sprite._position)
            sprite._internal.vertexArrays.setOffsets(sprite._internal.slots, 
                                            sprite._baseOffsets)
        
    def add(self, thing):
        """Add a sprite or iterable container of sprites to the group
        Arguments:
         container: the sprite or container to add"""
        if isinstance(thing, Sprite):
            self.addSprite(thing)
        else:
            try:
                for sprite in thing:
                    self.addSprite(sprite)
            except TypeError:
                raise Exception('Unsure how to add %s to SpriteGroup' 
                                 % str(thing))

