"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.sprites.shader

Provides the sprite shader
"""

from __future__ import division
from weakref import WeakSet
from pygl3display.twodshaders import TwoDShaderCommon, TwoDShader
from pygl3display.texture.twod import TwoDTextureCommon
from pygl3display.sprites.spritecommon import SpriteCommon

class SpriteShader(TwoDShader):
    """Shader for Two-D sprite shaders"""
    
    def __init__(self, atlasClass, defaultFilter = 'bilinear', 
                 cameraObj = None):
        """Initialise the 2d sprite shader
        Args:
         atlasClass: Class for texture atlases
         defaultFilter: the default filter to apply
         cameraObj: A TwoDCamerashader object for camera state. Default to 
                    shared camera state."""
        if cameraObj is None:
            cameraObj = TwoDShaderCommon.defaultCameras
        
        vertexProg = """
            vec2 applyTransform(in vec2 pos, in vec2 origin, in vec2 offset, 
                       in float xScale, in float yScale, in float rotation);
            vec2 applyCameras(in vec4 camera, in vec2 pos);
            attribute vec4 pos;   // x,y = position, z,w offset 
            attribute vec4 texCoords; // x,y = x0, y0, z,w = x1-x0, y1-y0
            attribute vec4 texRatios; // x,y = texture ratio
            attribute vec4 color; 
            attribute vec4 rotozoom; // x = scaleX, y = scaleY, z = rotation
            attribute vec4 cameras1;
            attribute vec4 cameras2;
            attribute vec4 fragCameras;
            attribute vec4 filters;
            uniform mat4 projMatrix;
            varying vec4 v2fColor;
            varying vec4 v2fTex;
            varying vec4 v2fTilingPos;
            varying vec4 v2fCameras;
            varying vec4 v2fFilters;
            void main()
            {   
                vec2 transform = applyTransform(pos.xy, vec2(0, 0), 
                                 pos.zw, rotozoom.x, rotozoom.y, rotozoom.z);
                transform = applyCameras(cameras1, transform);
                transform = applyCameras(cameras2, transform);
                gl_Position = projMatrix * vec4 (transform.xy, 0, 1);
                v2fTilingPos.zw = gl_Position.xy;
                v2fTex = texCoords; 
                v2fTilingPos.xy = texRatios.xy;
                v2fColor = color;
                v2fCameras = fragCameras;
                v2fFilters = filters;
            }"""

        fragmentProg = """
            vec4 applyFragCameras(in vec4 cameras, in vec4 fragColor);
            bool drop(in vec2 pos);
            uniform sampler2D tex;
            uniform float texTexel;
            uniform float texSize;
            varying vec4 v2fColor;
            varying vec4 v2fTex;
            varying vec4 v2fTilingPos;
            varying vec4 v2fCameras;
            varying vec4 v2fFilters; // x = type, yzw = params

            void main()
            {   
                if (drop(v2fTilingPos.zw)) {discard;} 
                vec2 texP = 
                   vec2(v2fTex.x + mod((v2fTilingPos.x * v2fTex.z), v2fTex.z),
                        v2fTex.y + mod((v2fTilingPos.y * v2fTex.w), v2fTex.w));
                vec4 texColor;
                if (v2fFilters.x == 0.0) {
                    texColor = texture2D(tex, texP);
                }
                else if (v2fFilters.x == 1.0) {
                    vec4 tl = texture2D(tex, texP);
                    vec4 tr = texture2D(tex, texP + vec2(texTexel, 0));
                    vec4 bl = texture2D(tex, texP + vec2(0, texTexel));
                    vec4 br = texture2D(tex, texP + vec2(texTexel , texTexel));
                    vec2 mixVals = fract( texP * texSize );
                    vec4 texColorT = mix( tl, tr, mixVals.x );
                    vec4 texColorB = mix( bl, br, mixVals.x );
                    texColor = mix(texColorT, texColorB, mixVals.y);
                }
                
                gl_FragColor = applyFragCameras(v2fCameras, 
                              v2fColor * texColor);
            }"""
        
        super(SpriteShader, self).__init__(
            vertexProgs= [TwoDShaderCommon.applyTransformSource, 
            cameraObj.shaderSource, vertexProg], 
            fragmentProgs=[TwoDShaderCommon.dropSource, 
            cameraObj.fragShaderSource, fragmentProg])
            
        self.atlases = WeakSet()
        self.defaultFilter = defaultFilter
        self.findAtlasSize()
        self.atlasClass = atlasClass
        
    def setAtlasClass(self, cls):
        """Set the class of texture atlas to use"""
        self.atlasClass = cls
        
    def setDefaultFilter(self, filter):
        """Set the default filter for sprites on this shader"""
        self.defaultFilter = filter
        
    def recreate(self):
        """Recreate all atlases in the shader - note that some atlases may
        not contain the same information (those which contain information
        only existing on the GPU)"""
        for atlas in self.atlases: 
            atlas.destroy()
        for atlas in self.atlases: 
            atlas.recreate()
        
    def findAtlasSize(self):
        """Function to determine maximum atlas size"""
        self.atlasSize = TwoDTextureCommon.findTextureSize()

    def makeNewAtlas(self):
        """Create a new atlas of maximum possible size"""
        self.findAtlasSize()
        temp = self.atlasClass(size=self.atlasSize, shader = self)
        self.atlases.add(temp)  
        return temp

    def getBox(self, width, height, image = None, wrapped = False):
        """Get a box in a texture atlas.
        Arguments:
         width  : width needed
         height : height needed
         image  : a Image object; if not specified, create a new Image 
        """
        if width >= self.atlasSize or height >= self.atlasSize:
            raise CannotAllocateError('Cannot allocate area of size ' +
                                      str(width) + 'x' + str(height))
        x = None
        for atlas in self.atlases:
            x = atlas.getBox(width, height, image, wrapped)
            if x is not None:
                break
                
        if x is None:
            x = self.makeNewAtlas()
            x = x.getBox(width, height, image, wrapped)
        return x        

SpriteCommon.setDefaultShaderClass(SpriteShader)

class CannotAllocateError(Exception):
    """Exception raised when a rectangle cannot be allocated in any atlas"""
