"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3).

pygl3display.sprites.sprite

This file provides "oldschool style" 2d Sprite specific features.

Constants:

topLeftOffsets, topRightOffsets, bottomLeftOffsets, bottomRightOffsets
centerOffsets:
  These offsets, if passed to a sprite as the offsets argument with
  the offsets given as relative will cause the named corner (or middle) 
  to be positioned at (0, 0)
"""

from __future__ import division
from pygl3display.twodbases import TwoDThing

bottomLeftOffsets = (0, 0), (1, 0), (1, -1), (0, -1)
bottomRightOffsets = (-1, 0), (0, 0), (0, -1), (-1, -1)
topLeftOffsets = (0, 1), (1, 1), (1, 0), (0, 0)
topRightOffsets = (-1, 1), (0, 1), (0, 0), (-1, 0)
centerOffsets = (-0.5, 0.5), (0.5, 0.5), (0.5, -0.5), (-0.5, -0.5)


class Sprite(TwoDThing):
    """A Sprite is primarily an Image and the information needed
    to draw it to screen."""
    
    filters = {'nearest': 0, 'bilinear': 1}
    dataSlotsReq, renderSlotsReq, pattern = 4, 6, [0, 1, 2, 0, 2, 3]
    
    def __init__(self, image, position=(0, 0), layer=0, target=None,
                 flipx=False, flipy=False,
                 scalex=1, scaley=1, scale=None, rotation=0, visible=True,
                 tiling=(1, 1), filter=None,
                 color=(255, 255, 255, 255),
                 offsets=topLeftOffsets, offsetsRel=True, 
                 cameras=None, fragCameras=None,
                 renderer=None, rslots = None,
                 vertexArrays = None):
        """Initialise the Sprite. Arguments (all optional):
           image: a Image to use as the sprites image.
           position: The (x, y) coordinate for the sprite.
                     Default (0, 0)
           offsets: A list/tuple of the offsets of the four corners of the 
                    sprite from position. Defaults to topLeftOffsets.
           offsetsRel: If True, multiply each of the offsets by the
                       sprites (width, height). This is needed for the preset
                       values of offsets. Defaults to True
           layer: The layer of the sprite. Default 0
           flipx: If the surface should be initially flipped X-axis. 
                  Default False
           flipy: If the surface should be initially flipped Y-axis. 
                  Default False
           visible: If the sprite should be visible. Default True
           filter: The filtering the sprite should use. 
                   Default to value from SpriteCommon
           rotation: The rotation of the sprite in radians. Default 0
           scalex: The x scaling of the sprite. Default 1.
           scaley: The y scaling of the sprite. Default 1.
           scale: The scale of the sprite. Default (not specified). 
                  Overrides scalex and scaley.
           tiling: If given, sprite will be tiled the number of time specified
                   the x and y directions.
                   NOTE1: If using Linear Filtering, the image should be loaded
                   as tiled or you will get a transparent line problem.
           color: The coloration of the sprite. Uses normal pygame RGBA color
                  values. Each pixel in the image is "multiplied" by this
                  color, so it's very useful for doing fading with the alpha
                  component. Default White with full opacity (255,255,255,255)
           cameras: A list of length < 8 or a dict with keys from a subset of 
                    1-8. In each case, values should be integers or camera
                    proxies. If a list, the first len(cameras) of the sprite 
                    will be set to the cameras specified. If a dict, the keys
                    are used to determine which cameras are set.
           fragCameras: as cameras, but length 4 and for fragment cameras
           renderer, rslots, vertexArrays, slots: 
                If all are specified, use the given renderer etc. 
                For use internally be makeAdjacentSprites; not recommended for
                external use.
        """        
        if image is not None and renderer is None:
            renderer = image.atlas
        
        super(Sprite, self).__init__(
            type(self).dataSlotsReq, type(self).renderSlotsReq, 
            type(self).pattern, renderer, layer, target)
        self.setVisibility(visible)
        
        if (renderer is not None and rslots is not None and 
            vertexArrays is not None):
            #self._intl.setRenderer(renderer)
            self._intl.lockRenderer(True)
        
        self._flipx = False
        self._flipy = False
        
        self.setImage(image)

        if scale is None:
            self.setScaleX(scalex)
            self.setScaleY(scaley)
        else:
            self.setScale(scale)
            
        if cameras is not None:
            self.setCameras(cameras)
        if fragCameras is not None:
            self.setFragCameras(fragCameras)
            
        if filter is None:
            if self._intl.renderer is not None:
                filter = self._intl.renderer.getDefaultFilter()
            else:
                filter = 'bilinear'
        
        self.setTiling(tiling, updateSize=False)    
        self.setOffsets(offsets, relative=offsetsRel)
        self.setColor(color)
        self.setPosition(position)
        self.setRotation(rotation)
        self.flipX(flipx)
        self.flipY(flipy)
        self.setFilter(filter)
                
    def setTiling(self, tiling, updateSize=True):
        """Set the tiling on the sprite.
        Arguments:
           tiling: Tile the sprite the given amount of time in the x and y
                   directions.
                   e.g. (2, 1) tiles the sprite twice in the x direction.
           updateSize: If True, the size of the sprite will be changed to
                       allow the tiles to be full size. If False, the size
                       of the sprite will remain the same.
                       
           IMPORTANT: If using filtering, the image should be loaded
           as tiled or you will get a transparent line problem.
        """
        self._tiling = tiling
        self._intl.vertexArrays['TexRatio'][self._intl.slots] = \
            ((0, tiling[1]), (tiling[0], tiling[1]), (tiling[0], 0), (0, 0))
        if updateSize:
            self.setOffsets(self._unmodOffsets, self._originRel)
            
    def setFilter(self, filter):
        """Sets the filtering on the sprite.
           Filter should be from the following table:
           0 / 'nearest' : Use nearest neighbour
           1 / 'bilinear': Use bilinear filtering
        """
        self._filter = filter
        if filter in type(self).filters:
            filter = type(self).filters[filter]
        self._intl.vertexArrays['FilterType'][self._intl.slots] = filter
        
    def setOffsets(self, offsets, relative=True):
        """Set the coordinates of each of the sprites coordinates
        Arguments:
         offsets: A list containing the top-left, top-right, 
                  bottom-right, bottom-left coordinates of the sprite
         relative: If true, multiple each coordinate tuple by current image
                   width and height"""
        if relative is True:
            width, height = self._width, self._height
            baseOffsets = [(width*ox, height*oy) for ox, oy in offsets]
        else:
            baseOffsets = offsets
        
        tilesX, tilesY = self._tiling
        self._baseOffsets = [(ox * tilesX, oy * tilesY) 
                              for ox, oy in baseOffsets]
        
        self._unmodOffsets = offsets
        self._originRel = relative
        self._intl.vertexArrays['Offset'][self._intl.slots] = self._baseOffsets
                
    def setImage(self, image):
        """Set the image of the sprite
        Arguments:
         image: An Image (or subclass, eg. Surface).
        """
        self._image = image
        if image is not None:
            self._width, self._height = self._image.getSize()
            self._intl.setRenderer(self._image.atlas)
            self._intl.vertexArrays['TexCoord'][self._intl.slots] = \
                                                            self._image.coords
        else:
            self._width, self._height = 0, 0
            self._baseOffsets = ((0, 0), (0, 0), (0, 0), (0, 0))
            self._unmodOffsets = self._baseOffsets
            self._intl.vertexArrays['Offset'][self._intl.slots] = \
                                                            self._baseOffsets
        
    def flipX(self, flipped = None):
        """Flip the sprite horizontally"""
        if flipped != self._flipx:
            self._flipx = not self._flipx
            self._baseOffsets = [self._baseOffsets[3], self._baseOffsets[2], 
                                self._baseOffsets[1], self._baseOffsets[0]]
            self._intl.vertexArrays['Offset'][self._intl.slots] = \
                                                        self._baseOffsets
            
    def flipY(self, flipped = None):
        """Flip the sprite vertically"""
        if flipped != self._flipy:
            self._flipy = not self._flipy
            self._baseOffsets = [self._baseOffsets[1], self._baseOffsets[0], 
                                self._baseOffsets[3], self._baseOffsets[2]]
            self._intl.vertexArrays['Offset'][self._intl.slots] = \
                                                         self._baseOffsets
            

def makeAdjacentSprites(images, layer, target = None, spriteclass=Sprite):
    """Make some sprites which are guanteed to be adjacent within the layer.
    Notes: All images must be on the same atlas. This can be accomplished by
    subsurfaces. None of the sprites may change layer."""
    if not images: 
        return []
    atlas = images[0].atlas
    if not all([image.atlas is atlas for image in images]):
        raise Exception('For adjacent sprites, all images must ' 
                        'be on the same atlas')
    vertexArrays = atlas.getVertexArrays(4 * len(images))
    slots = atlas.getRenderSlotBlocks(6, len(images), layer, target, 
                                      vertexArrays) 
    return [spriteclass(image=image, layer=layer, renderer=atlas, 
                        rslots=rslots, vertexArrays=vertexArrays)
            for image, rslots in zip(images, slots)]
