"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.twodtextureconvertors

Provides some utility functions that allow simple conversion of a 2D texture
to TextureAtlas, Image or Sprite.
"""
from __future__ import division

from pygl3display.sprites.textureatlas import TextureAtlas2d
from pygl3display.sprites.sprite import Sprite
import weakref

class ConverterClass(object):
    """A class which provides the facilities to convert textures to Atlases,
    Images or Sprites"""
    
    def __init__(self):
        """Initialise the convertor"""
        self.textureAtlases = weakref.WeakKeyDictionary()
        
    def textureAsAtlas(self, texture, atlas=None):
        """Make a texture atlas of the texture"""
        if atlas is None:
            if (texture in self.textureAtlases and 
            self.textureAtlases[texture]()):
                return self.textureAtlases[texture]()
            else:
                newAtlas = TextureAtlas2d(size = texture.size, 
                                          texture = texture, rects = [])
                self.textureAtlases[texture] = weakref.ref(newAtlas)
                return newAtlas
        else:
            oldTex = atlas.texture
            if oldTex in self.textureAtlases:
                del self.textureAtlases[oldTex]
            self.textureAtlases[texture] = weakref.ref(atlas)
            atlas.setTexture(texture)
        
    def textureAsImage(self, texture, topleft = (0, 0), bottomright = None):
        """Make an image of the texture"""
        if bottomright is None:
            bottomright = (texture.size, texture.size)
        texatlas = self.textureAsAtlas(texture)
        size = texture.size
        left, top = ((topleft[0] // size), (topleft[1] // size))
        right, bottom = (int(bottomright[0] / size), int(bottomright[1] / size))
        return texatlas.getAtlasImage(top, bottom, left, right)
        
    def textureAsSprite(self, texture, topleft = (0, 0), bottomright = None, 
                        sprite=None):
        """Make a sprite of the texture; if you want to render the texture
        to screen, this is the function to use."""
        if sprite is None:
            image = self.textureAsImage(texture, topleft, bottomright)
            return Sprite(image, flipx=True)
        else:
            self.textureAsAtlas(texture, sprite._image.atlas)
            return sprite

Converter = ConverterClass()
