"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.texture.base

Provides an abstract base for textures
"""

from pygl3display.uniforms.types import FloatUniform, IntUniform
import weakref
from pygl3display.gl import ( TexUnitsDict,
    glBindTexture, glTexParameteri, glGetTexImage, glActiveTexture,
    glDeleteTexture, glGenTextures,
    GL_UNSIGNED_BYTE, GL_MAX_TEXTURE_COORDS, GL_TEXTURE_MIN_FILTER, 
    GL_TEXTURE_MAG_FILTER, GL_NEAREST, GL_FLOAT, GL_RED, GL_GREEN, GL_BLUE, 
    GL_ALPHA, GL_RGB, GL_BGR, GL_RGBA, GL_BGRA, GL_LUMINANCE, 
    GL_LUMINANCE_ALPHA)
    
    
COLORMASKS = {'r': GL_RED, 'g': GL_GREEN, 'b': GL_BLUE, 'a': GL_ALPHA, 
    'rgb': GL_RGB, 'bgr': GL_BGR, 'rgba': GL_RGBA, 'bgra': GL_BGRA, 
    'l': GL_LUMINANCE, 'la': GL_LUMINANCE_ALPHA}


class TextureCommonClass(object):
    """The shared state between all Textures"""
    
    def __init__(self):
        """Initialise the shared state"""
        self.texUnits = TexUnitsDict()
        self.usage = [x for x in xrange(self.numberOfTextureUnits)]
        self.tex2tu = {}
        self.tu2tex = {}
        self.activetu = 0
        self.texNames = set()
        self.boundTexNames = {}
    
    def genTexture(self, user):
        """Generate a single texture
        Args:
          user: The object that will hold the texture name"""
        return self.genTextures(user, 1)[0]
    
    def genTextures(self, user, n):
        """Generate textures which are dealloced when the user is GC'd
        Args:
          user: The object that will hold the texture name
          n: The number of texture names to generate."""
        ret = []
        while self.texNames and n > 0:
            ret.append(self.texNames.pop())
            n -= 1
        if n > 0:
            ret.extend(glGenTextures(n))
        userRef = weakref.ref(user, self.freeTexNames)
        self.boundTexNames[userRef] = ret
        return ret

    def freeTexNames(self, ref):
        """Unbind a texture name"""
        self.texNames.update(self.boundTexNames[ref])
        del self.boundTexNames[ref]

    def __getattr__(self, key):
        """Load the number of texture units on demand"""
        if key == 'numberOfTextureUnits':
            self.numberOfTextureUnits = GL_MAX_TEXTURE_COORDS
            # This should probably be moved to hacks after things have been
            # tracked down to specific drivers etc.
            if self.numberOfTextureUnits > 16: 
                self.numberOfTextureUnits = 16
            return self.numberOfTextureUnits
        else:
            super(TextureCommonClass, self).__getattribute__(key)
        
    def activateTU(self, tu):
        """Activates a texture unit. tu should be an 
        integer < GL_MAX_TEXTURE_COORDS."""
        if self.activetu != tu:
            glActiveTexture(self.texUnits[tu])
            self.activetu = tu
            
    def bindTextures(self, textures):
        """Checks to see if given textures could be bound simultaneously;
        binds them all if possible, otherwise raises an exception"""
        if len(textures) < self.numberOfTextureUnits:
            for texture in textures:
                self.bindTexture(texture)
        else:
            raise Exception('Only %i texture units available' 
                             % self.numberOfTextureUnits)
    
    def setTexTU(self, texture, tu):
        """Sets a texture to use a specific texture unit"""
        self.usage.remove(tu)
        self.usage.append(tu)
        self.activateTU(tu)
        glBindTexture(texture.textureType, texture.surface)
        texture.setTU(tu)
        texid = texture.surface
        if self.tu2tex.get(tu, texid) != texid:
            oldtex = self.tu2tex[tu]
            if oldtex in self.tex2tu: 
                del self.tex2tu[oldtex]
        self.tu2tex[tu] = texid
        self.tex2tu[texid] = tu
        
    def bindTexture(self, texture):
        """Ensure a texture is bound to a texture unit"""
        texID = texture.surface
        if texID in self.tex2tu:
            tu = self.tex2tu[texID]
        else:
            tu = self.usage[0]
        self.setTexTU(texture, tu)
    
TextureCommon = TextureCommonClass()

class Texture(object):
    """An abstract hardware accelerated texture"""
    
    def initTexture(self, size, data, textureType, initialiser):
        """Initialise the texture
        Args:
            size: Size of the texture
            textureType: 
            initiliser: a lambda function to initialise the texture
        """
        self.size = size
        self.data = data
        self.textureType = textureType
        self.initialiser = initialiser
        
        self.tuUniform = IntUniform()
        self.tu = 0
        self.sizeUniform = FloatUniform(self.size)
        self.texelUniform = FloatUniform(1.0 / self.size)

        self.surface = TextureCommon.genTexture(self)
        TextureCommon.bindTexture(self)
        self.initialiser()
        glTexParameteri(textureType, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(textureType, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
                
    def setTU(self, tu):
        """Sets the texture unit the texture was last bound to"""
        if tu != None:
            self.tuUniform.uploadData(tu)
        self.tu = tu
                
    def getTexture(self):
        """Returns the gl Texture identifier"""
        return self.surface
        
    def destroy(self):
        """Destroy the gl texture"""
        if self.surface is not None: 
            glDeleteTexture(self.surface)
        self.surface = None
        
    def recreate(self):
        """Recreates the texture"""
        self.initTexture(self.size, self.data, self.textureType, 
                         self.initialiser)
        
    def bindTexture(self, layer=0):
        """Callback to bind the texture to the texture unit"""
        TextureCommon.bindTexture(self)
        
    def readPixels(self, mask = GL_RGBA):
        """Dump pixels from texture to a 3d numpy array. This is not
        really ideal, but alternatives are slow - this will be subject
        to change should I implement my own GL bindings.
        Arguments:
            mask: Either an OpenGL mask or an easier to use string
            from COLORMASKS"""
        mask = COLORMASKS.get(mask, mask)
        self.bindTexture()
        TextureCommon.activateTU(self.tu)
        nppixels = glGetTexImage(self.textureType, 0, mask, GL_FLOAT)
        return nppixels
        
    def readPixelsUB(self, mask = GL_RGBA):
        """Dump pixels as unsigned bytes. This returs a string, and is
        likely faster than anything readPixels will do. The returned string
        is also appropriate for uploading somewhere else. Of course,
        accuracy is lost."""
        mask = COLORMASKS.get(mask, mask)
        self.bindTexture()
        TextureCommon.activateTU(self.tu)
        pixels = glGetTexImage(self.textureType, 0, mask, GL_UNSIGNED_BYTE)
        return pixels
        

