"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.twodbases

Provides base classes that 2d implementations can derive from
"""


class Internal(object):
    """Encapsulate the data storage requirements of a TwoDThing"""
    def __init__(self, dataSlotsReq, renderSlotsReq, pattern):
        """Initialise the Internal object"""
        self.__dataSlotsReq = dataSlotsReq
        self.__renderSlotsReq = renderSlotsReq
        self.__pattern = pattern
        self.renderer = None
        self.vertexArrays = None
        self.slots = None
        self.rslots = None
        self.rdata = None
        self.layer = None
        self.target = None
        self.render = True
        self.lockedRenderer = False
        
    def setRenderer(self, newRenderer):
        """Set the renderer this object uses"""
        if self.layer is None:
            raise Exception('No layer assigned')
        if self.renderer != newRenderer and not self.lockedRenderer:
            self.renderer = newRenderer
            self.vertexArrays = self.renderer.getVertexArrays(
                                    self.__dataSlotsReq)
            if self.slots is not None:
                self.slots = self.vertexArrays.copySlots(self.slots)
            else:
                self.slots = self.vertexArrays.getDataSlots(
                                    self.__dataSlotsReq)
            self.rslots = self.renderer.getRenderSlots(self.__renderSlotsReq, 
                            self.layer, self.target, self.vertexArrays)
            self.rdata = [self.slots[x] for x in self.__pattern]
            if self.render is True:
                self.rslots.upload(self.rdata)
                
    def lockRenderer(self, lock):
        """Locks the renderer that this object uses"""
        self.lockedRenderer = lock
            
    def setRender(self, render=True):
        """Set if this object should render to screen"""
        if self.render is not render:
            self.render = render
            if self.render is True:
                self.rslots.upload(self.rdata)
            else:
                self.rslots.upload([0] * len(self.rdata))
                
    def setLayer(self, layer):
        """Set the layer of this internal object"""
        if layer != self.layer and self.renderer is not None:
            self.rslots = self.renderer.getRenderSlots(self.__renderSlotsReq, 
                                    layer, self.target, self.vertexArrays)
            self.rslots.upload(self.rdata)
        self.layer = layer
        
    def setTarget(self, target):
        """Set the render target of this internal object"""
        if (target != self.target and self.renderer is not None and 
        self.layer is not None):
            self.rslots = self.renderer.getRenderSlots(self.__renderSlotsReq, 
                                    self.layer, target, self.vertexArrays)
            self.rslots.upload(self.rdata)
        self.target = target
        
        
class TwoDThing(object):
    """A TwoDThing is anything that should have access to this common set of
    functions."""
    def __init__(self, dataSlotsReq, renderSlotsReq, pattern, renderer, 
                 layer, target):
        """Initialises all variables for the twodthing"""
        self._intl = Internal(dataSlotsReq, renderSlotsReq, pattern)
        self.setLayer(layer)
        self.setTarget(target)
        self._intl.setRenderer(renderer)
        
    def setScale(self, scale):
        """Set the scale of the thing"""
        self._intl.vertexArrays['ScaleX'][self._intl.slots] = scale
        self._intl.vertexArrays['ScaleY'][self._intl.slots] = scale
        self._scalex = scale
        self._scaley = scale
        
    def setScaleX(self, scale):
        """Set the X scale of the thing; scaling applied before rotation"""
        self._intl.vertexArrays['ScaleX'][self._intl.slots] = scale
        self._scalex = scale

    def setScaleY(self, scale):
        """Set the Y scale of the thing; scaling applied before rotation"""
        self._intl.vertexArrays['ScaleY'][self._intl.slots] = scale
        self._scaley = scale
        
    def setCamera(self, n, camera):
        """Set the nth camera applied to the thing"""
        if n < 4:
            cameraArray = self._intl.vertexArrays['Camera1'] 
        else:
            cameraArray = self._intl.vertexArrays['Camera2']
            n -= 4 
        cameraArray[n][self._intl.slots] = int(camera)
        
    def setFragCamera(self, n, camera):
        """Set the nth fragment camera applied to the thing"""
        self._intl.vertexArrays['FragCamera'][n][self._intl.slots] = \
                                                             int(camera)
        
    def setCameras(self, cameras):
        """Set multiple cameras in one function call.
        If the cameras argument is a list or tuple, the first len(cameras) 
        cameras of the thing are set. If the list has length > 8, only the 
        first 8 are used. 
        If the cameras argument is a dict, it should map a subset of the
        values 0-8 to cameras. Keys not in the range 0-8 are ignored"""
        if isinstance(cameras, (list, tuple)):
            for x in xrange(min(len(cameras), 8)):
                self.setCamera(x, cameras[x])
        elif isinstance(cameras, dict):
            for x in cameras:
                if 0 <= x < 8:
                    self.setCamera(x, cameras[x])
        else:
            raise ValueError('cameras should be specified as a list or dict')
            
    def setFragCameras(self, cameras):
        """Set multiple fragment cameras in one function call.
        If the cameras argument is a list or tuple, the first len(cameras) 
        cameras of the thing are set. If the list has length > 4, only the 
        first 4 are used. 
        If the cameras argument is a dict, it should map a subset of the
        values 0-3 to cameras. Keys not in the range 1-8 are ignored"""

        if isinstance(cameras, (list, tuple)):
            for x in xrange(min(len(cameras), 4)):
                self.setFragCamera(x, cameras[x])
        elif isinstance(cameras, dict):
            for x in cameras:
                if 0 <= x < 4:
                    self.setFragCamera(x, cameras[x])
        else:
            raise ValueError('cameras should be specified as a list or dict')
            
    def setColor(self, color):
        """Set the coloration of the thing"""
        self._color = color
        self._intl.vertexArrays['Color'][self._intl.slots] = color

    def setRotation(self, rotation):
        """Set the rotation of the thing"""
        self._rotation = rotation
        self._intl.vertexArrays['Rotation'][self._intl.slots] = rotation

    def setPosition(self, position):
        """Set the position of the thing"""
        self._position = position
        self._intl.vertexArrays['Position'][self._intl.slots] = position

    def setVisibility(self, visibility):
        """Set the visibility of the thing"""
        self._visible = True if visibility else False
        self._intl.setRender(self._visible)
                
    def setLayer(self, layer):
        """Set the layer of the thing"""
        self._intl.setLayer(layer)
        
    def setTarget(self, target):
        """Set the target of the thing"""
        self._intl.setTarget(target)

    def setColors(self, colors):
        """Set all the vertice colors of the thing
        Args:
          colors: A list of coordinates for the vertices of the lines
          WARNING: This trusts that the number of colors is the same as
          the number of vertices in the lines, otherwise this will fail
          with a cryptic error message"""
        self._colors = colors
        self._intl.vertexArrays['Colors'][self._intl.slots] = self._colors
        
    def setVertexColor(self, n, color):
        """Set the color of the nth vertex in the thing
        Args:
          n: The number of the vertex to set
          color: The new color to set the vertex to
        """
        self._colors[n] = color
        self._intl.vertexArrays['Colors'][self._intl.slots] = self._colors

    def setVertices(self, vertices):
        """Set all the vertices of the thing
        Args:
          vertices: A list of coordinates for the vertices of the lines
          WARNING: This trusts that the number of vertices is the same as
          previous, otherwise this will fail with a cryptic error message"""
        self._vertices = vertices
        self._intl.vertexArrays['Offset'][self._intl.slots] = self._vertices
                
    def setVertex(self, n, pos):
        """Set the position of the nth vertex in the thing
        Args:
          n: The number of the vertex to set
          pos: The new position to set the vertex to
        """
        self._vertices[n] = pos
        self._intl.vertexArrays['Offset'][self._intl.slots] = self._vertices

class TwoDBaseCommonClass(object):
    """Common state of TwoDThings"""
    def __getattr__(self, attrib):
        """Enable lazy creation of the default shader"""
        if attrib == 'defaultShader':
            self.makeDefaultShader()
            return self.defaultShader
        else:
            return self.__getattribute__(attrib)
    
    
if __name__ == '__main__':
    for attr in TwoDThing.__dict__.values():
        if isinstance(attr, type(lambda: 0)):
            print(attr.__name__ + u' = TwoDThing.' + attr.__name__)
            
        
