"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.uniforms.base

Provides basic representation of Uniforms
"""

import weakref
from pygl3display.glcommoninst import GLCommon


class Uniform(object):
    """Represents a Uniform"""
    def __init__(self, data, func, args = None, defaultName = None):
        """Initialise the Uniform"""
        self.data = data
        self.func = func
        self.args = args if args is not None else []
        self.names = weakref.WeakKeyDictionary()
        self.defaultName = defaultName
        self.shaderBindArgs = weakref.WeakKeyDictionary()
    
    def assignName(self, name, shader):
        """Informs the uniform of a name to use on a given shader"""
        self.names[shader] = name
    
    def dirty(self):
        """Mark the uniform as dirty"""
        for shader in self.shaderBindArgs:
            self.shaderBindArgs[shader][0] = True
            
    def uploadData(self, *args):
        """Upload new data to the uniform"""
        self.data = list(args)
        self.dirty()
        
    def genBindArgs(self, shader):
        """Generate arguments for binding the uniform on a shader"""
        
        if shader in self.names or self.defaultName is not None:
            name = self.names.get(shader, self.defaultName)
            location = shader.uniformLocs[name]
            self.shaderBindArgs[shader] = [True, [location] + self.args]
        else:
            raise Exception(u'Uniform does not have a name for given shader')
        
    def bind(self, shader, force):
        """Bind the uniform"""
        if shader not in self.shaderBindArgs:
            self.genBindArgs(shader)
        shaderBindArgs = self.shaderBindArgs[shader]
        if shaderBindArgs[0] or force:
            GLCommon.bindUniform(self.func, shaderBindArgs[1] + self.data, 
                                force=shaderBindArgs[0])
            shaderBindArgs[0] = False
              
class ProxyUniform(object):
    """A proxy of a uniform"""
    def __init__(self):
        """Initialise variables"""
        self.__uniform = None
        
    def __getattr__(self, key):
        """Return the item from the proxied uniform"""
        return getattr(self.__uniform, key)
        
    def proxy(self, uniform):
        """Set the uniform to proxy"""
        if self.__uniform != uniform:
            self.__uniform = uniform
            uniform.dirty()
