"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.uniforms.types

Provides specific classes for the different types of OpenGL Uniforms
"""
from __future__ import division

from pygl3display.gl import (glUniform1f, glUniform1i, glUniform4fv,
    glUniformMatrix4fv)

from pygl3display.uniforms.base import Uniform

def _defaultVectorValueLength(value, length, perLength):
    """Work out the length of a given vector"""
    if length is None:
        if value is not None:
            length = len(value) // perLength
        else:
            length = 1
    value = value if value is not None else [0] * (length * perLength)
    return value, length

class FloatUniform(Uniform):
    """A float uniform"""
    def __init__(self, value=0.0, defaultName=None):
        """Initialise the float uniform.
        Args:
          value: The value of the uniform
          defaultName: The default name the uniform should be bound to in GLSL
                       (can be overridden on a per shader basis"""
        super(FloatUniform, self).__init__([value], glUniform1f, [], 
                                           defaultName)
        
class IntUniform(Uniform):
    """An integer uniform"""
    def __init__(self, value=0, defaultName=None):
        """Initialise the int uniform.
        Args:
          value: The value of the uniform
          defaultName: The default name the uniform should be bound to in GLSL
                       (can be overridden on a per shader basis"""
        super(IntUniform, self).__init__([value], glUniform1i, [], 
                                         defaultName)

class Vec4VectorUniform(Uniform):
    """A uniform containing a vector of Vec4s"""
    def __init__(self, value=None, length=None, defaultName=None):
        """Initialise the Vec4 Vector uniform.
        Args:
          value: The value of the uniform, as a list; defaults to 0's
          length: The number of vec4s in the vector; defaults to 1
          defaultName: The default name the uniform should be bound to in GLSL
                       (can be overridden on a per shader basis"""
        value, length = _defaultVectorValueLength(value, length, 4)
        super(Vec4VectorUniform, self).__init__([value], glUniform4fv, 
                                                [length], defaultName)
        
class Matrix4x4VectorUniform(Uniform):
    """A uniform containing a vector of Mat4x4s"""
    def __init__(self, value=None, length=None, transpose=False, 
                 defaultName=None):
        """Initialise the Matrix 4x4 Vector uniform.
        Args:
          value: The value of the uniform, as a list; defaults to 0's
          length: The number of mat4x4s in the vector; defaults to 1
          defaultName: The default name the uniform should be bound to in GLSL
                       (can be overridden on a per shader basis"""
        value, length = _defaultVectorValueLength(value, length, 16)
        super(Matrix4x4VectorUniform, self).__init__(
            [value], glUniformMatrix4fv, [length, transpose], defaultName) 