"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.vertexarrays.accessors

Provides accessors onto vertexarrays
"""

from itertools import izip
from collections import Iterable

class AccessorMeta(type):
    """Metaclass for accessors; each type of array gets its own class"""
    def setup(cls, arrayName, blockSize, itemSize, offset):
        """Setup constants for this type of array"""
        cls.arrayName = arrayName
        cls.blockSize = blockSize
        cls.itemSize = itemSize
        cls.offset = offset
        
class Accessor(object):
    """Provides easy indexed access to vertex arrays"""
    __metaclass__ = AccessorMeta
    
    def __init__(self, arrays, dirtyArrays, dirtyDyn):
        """Initialise the accessor"""
        self.array = arrays[type(self).arrayName]
        self.dirtyArrays = dirtyArrays
        self.dirtyDyn = dirtyDyn[type(self).arrayName]
        # Inherit the values from the metaclass: 
        # TODO: A more elegant way of achieving this.
        self.arrayName = type(self).arrayName
        self.blockSize = type(self).blockSize
        self.itemSize = type(self).itemSize
        self.offset = type(self).offset

    def __setitem__(self, indx, val):
        """Basic handling for setting values with slices"""
        if isinstance(indx, slice):
            for index, value in \
            izip(xrange(indx.start, indx.stop, indx.step), val):
                self._setSingleItem(index, value)
        elif isinstance(indx, Iterable):
            for index, value in izip(indx, val):
                self._setSingleItem(index, value)
        else:
            self._setSingleItem(indx, val)
            
    def _setSingleItem(self, indx, val):
        """Virtual method that must be overridden"""
        raise Exception('Not implemented')
                    
class AccessorSingle(Accessor):
    """Overrides setitem so that it sets multiple indices to the same value"""
    
    def __setitem__(self, indx, val):
        """Handle setting batches of values with slices"""
        if isinstance(indx, slice):
            for index in xrange(indx.start, indx.stop, indx.step):
                self._setSingleItem(index, val)
        elif isinstance(indx, Iterable):
            for index in indx:
                self._setSingleItem(index, val)
        else:
            self._setSingleItem(indx, val)    

class Tuple(Accessor):
    """Provides functionality for setting using a tuple"""
        
    def __getitem__(self, indx):
        """Get tuples from vertex arrays"""
        if isinstance(indx, slice):
            return [self[x] for x in xrange(indx.start, indx.stop, indx.step)]
        elif isinstance(indx, Iterable):
            return [self[x] for x in indx]
        else:
            start = (self.blockSize * indx) + self.offset
            end = start + self.itemSize
            return self.array[start:end]
        
    def _setSingleItem(self, indx, val):
        """Set a single tuple item"""
        self.dirtyArrays.add(self.arrayName)
        start = (self.blockSize * indx) + self.offset
        end = start + self.itemSize
        self.array[start:end] = val
        
class TupleDyn(Tuple):
    """Provides functionality for setting a tuple array, 
    and updates dynamic dirty list"""
    
    def _setSingleItem(self, indx, val):
        """Set a single tuple item"""
        self.dirtyArrays.add(self.arrayName)
        start = (self.blockSize * indx) + self.offset
        end = start + self.itemSize
        self.dirtyDyn[0].add(start)
        self.dirtyDyn[1].add(end)
        self.array[start:end] = val
 
            
class TupleSingle(Tuple, AccessorSingle):
    """A Tuple that assigns a single value to multiple indices"""

class TupleSingleDyn(TupleSingle, TupleDyn):
    """A TupleSingle with support for dynamic arrays""" 
            
class Individual(Accessor):
    """Provides access to items which are individual values"""
        
    def __getitem__(self, indx):
        """Allows seeing items which are single values"""
        if isinstance(indx, slice):
            return [self[x] for x in xrange(indx.start, indx.stop, indx.step)]
        elif isinstance(indx, Iterable):
            return [self[x] for x in indx]
        else:
            return self.array[(self.blockSize * indx) + self.offset]
    
    def _setSingleItem(self, indx, val):
        """Sets a single value item"""
        self.dirtyArrays.add(self.arrayName)
        self.array[(self.blockSize * indx) + self.offset] = val
        
class IndividualDyn(Individual):
    """As Individual, with support for dynamic arrays"""
    
    def _setSingleItem(self, indx, val):
        """Sets a s single value item"""
        self.dirtyArrays.add(self.arrayName)
        actualIndex = (self.blockSize * indx) + self.offset
        self.dirtyDyn[0].add(actualIndex)
        self.dirtyDyn[1].add(actualIndex)
        self.array[actualIndex] = val
        
            
class IndividualSingle(Individual, AccessorSingle):
    """As Individual, but assigns multiple indices a single value"""

class IndividualSingleDyn(IndividualSingle, IndividualDyn):
    """As IndividualBatch, but with dynamic array support"""

class OffsetAccessorMeta(type):
    """Metaclass for OffsetAccessors"""
    
    def setup(cls, accessorClass):
        """Setup the OffsetAccessor class"""
        cls.accessorClass = accessorClass
        
class OffsetAccessor(object):
    """Creates accessors with offsets equal to the index given"""
    
    __metaclass__ = OffsetAccessorMeta
    
    def __init__(self, array, dirtyArrays, dirtyDyn):
        """Initialse the offsetaccessor"""
        self.accessorArgs = [array, dirtyArrays, dirtyDyn]
        self.accessors = {}
        
    def __iter__(self):
        """Support for iteration"""
        return self.accessors.__iter__()
        
    def __getitem__(self, indx): 
        """Return the appropriate accessor"""
        if indx not in self.accessors:
            self.accessors[indx] = type(self).accessorClass(*self.accessorArgs)
            self.accessors[indx].offset = indx
        return self.accessors[indx]