"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3).

pygl3display.vertexarrays.obj

Provides the vertex arrays objects; this is only a basic template, and a
specific version should be constructed using a VertexArraySpec
"""
from __future__ import division

import weakref
import itertools
from pygl3display.glcommoninst import GLCommon
from pygl3display.slices import Slices
from pygl3display.gl import (glBufferData, glBufferSubData, 
    GL_UNSIGNED_INT, GL_ARRAY_BUFFER,  GL_STATIC_DRAW, GL_DYNAMIC_DRAW, 
    cdata, sizeOf, offsetPointer, byref)
from pygl3display.genstore import Buffers
from pygl3display.vertexarrays.accessors import Accessor, OffsetAccessor

class VertexSlots(object):
    """Slots on a vertex arrays object"""
    def __init__(self, slots, vertexArrays):
        """Initialise the slots"""
        self.slots = slots
        self.vertexArrays = vertexArrays
        
    def __str__(self):
        """A pretty print str representation, for debug"""
        return "VertexSlots(%s, %s)" % (self.slots, self.vertexArrays)
        
    def __len__(self):
        """Number of slots"""
        return len(self.slots)
    
    def __iter__(self):
        """Iterate on the slots"""
        return iter(self.slots)
    
    def __getitem__(self, indx):
        """Get an item from the slots"""
        return self.slots.__getitem__(indx)
    
    
class VertexArrays(object):
    """Handles Vertex arrays, so that they can be shared amongst
    shaders as necessary. This class should not be initialised directly;
    instead, use VertexArraySpec to create a subclass which has all the
    appropriate setter methods already created."""

    
    def __init__(self, noSlots, arrays, accessors):
        """Initialise the vertex array container
        Arguments:
         noSlots: The total number of slots (i.e. length of the arrays)
         arrays : A Dictionary mapping array names (as in the shader program)
                  to a list containing the array itself as first element,
                  and as second element a list of the arguments to 
                  glVertexAttribPointer, without the first (location) or 
                  last (the actual array) argument
                  e.g. [someCTypesArray, [4, GL_FLOAT, GL_FALSE, 0]]
         accessors: a dictionary of accessors onto the vertex arrays
         """
        
        self.positions = Slices(0, noSlots)
        self.arrays = arrays
        self.arraysR = dict([(key, self.arrays[key][0]) 
                              for key in self.arrays])
        self.shaderBindArgs = weakref.WeakKeyDictionary()
        self.dirty = set()
        self.dirtyRange = dict([(key, (set(), set())) for key in self.arrays])
        self.noSlots = noSlots
        self.indicesType = GL_UNSIGNED_INT
        self.indicesCType = cdata['I']
        self.sizeOfIndices = 4
        self.accessors = {}
        for accessorName in type(self).accessorClasses:
            accessor = self.accessors[accessorName] = \
                                    type(self).accessorClasses[accessorName] 
            accessor(self.arraysR, self.dirty, self.dirtyRange)
        buffers = Buffers.get(self, len(arrays))
        self.arrayBuffers = dict(zip(arrays.keys(), buffers))
        self.vertexArrayRefs = {}
        
    def __getitem__(self, key):
        return self.accessors[key]

    def copySlots(self, slots):
        """Copy a slots object onto this vertex arrays if possible;
        if it is, return the new Slots object"""
        if isinstance(slots.vertexArrays, type(self)):
            slotsLen = len(slots.slots)
            newSlots = self.getDataSlots(slotsLen)
            for oslot, nslot in itertools.izip(slots, newSlots):
                for name in self.accessors:
                    accessor = self.accessors[name]
                    oaccessor = slots.vertexArrays.accessors[name]
                    if isinstance(accessor, Accessor):
                        accessor[nslot] = oaccessor[oslot]
                    elif isinstance(self.accessors[name], OffsetAccessor):
                        for offset in oaccessor:
                            accessor[offset][nslot] = oaccessor[offset][nslot]
                    else:
                        raise Exception('Unknown accessor type %s' 
                                        % str(type(accessor)))
            return newSlots
        else:
            raise Exception('Incorrect type')
            
        
    def canAllocDataSlots(self, no):
        """Returns True if the VertexArrays have no or more slots remaining"""
        return self.positions.canAllocate(no) 
        
    def getDataSlots(self, no):
        """Get slots for storing vertex information
        Arguments:
         no: The number of slots to get"""
        slots = self.positions.allocate(no)
        ret = VertexSlots(slots, self)
        slotsRef = weakref.ref(ret, self.releaseDataSlots)
        self.vertexArrayRefs[slotsRef] = slots
        return ret
        
    def releaseDataSlots(self, slotsRef):
        """Release slots to be reused
        Arguments:
         slots: The slots to be released"""
        if slotsRef in self.vertexArrayRefs:
            self.positions.release(self.vertexArrayRefs[slotsRef])
            del self.vertexArrayRefs[slotsRef]

class VertexArraysOnline(VertexArrays):
    """A Vertex Arrays class which is 'online' i.e. can be bound to
    a GL context etc."""
    def __init__(self, noSlots, arrays, accessors):
        super(VertexArraysOnline, self).__init__(noSlots, arrays, accessors)
        for key in self.arrayBuffers:
            bufferID = self.arrayBuffers[key]
            data = self.arraysR[key]
            mult = (sizeOf(data) // len(data))
            GLCommon.bindBuffer(GL_ARRAY_BUFFER, bufferID)
            bufferType = (GL_DYNAMIC_DRAW if self.arrays[key][2] 
                          else GL_STATIC_DRAW)
            glBufferData(GL_ARRAY_BUFFER, len(data) * mult, data, bufferType)
        
    def bind(self, force=False):
        """Bind the arrays as necessary, using the currently bound shader
        to determine locations"""
        shader = GLCommon.boundShader
        if shader not in self.shaderBindArgs:
            bindArgs = {}
            for key in self.arrays:
                if key in shader.arrayLocs:
                    bindArrayArgs = ((shader.arrayLocs[key],) + 
                        self.arrays[key][1] + (offsetPointer(0),))
                    bindArgs[key] = bindArrayArgs
            self.shaderBindArgs[shader] = bindArgs
        bindArgs = self.shaderBindArgs[shader]
        if GLCommon.boundVertexArrays is not self or force:
            GLCommon.boundVertexArrays = self
            bind = bindArgs
        else:
            bind = self.dirty

        for array in bind:
            bufferID = self.arrayBuffers[array]
            data = self.arraysR[array]
            dynamic = self.arrays[array][2]
            GLCommon.bindBuffer(GL_ARRAY_BUFFER, bufferID)
            if array in self.dirty:
                if dynamic:
                    dirtyElems = self.dirtyRange[array]
                    mn, mx = min(dirtyElems[0]), max(dirtyElems[1])
                    mult = (sizeOf(data) // len(data))
                    sz = (mx - mn + 1) * mult
                    ref = byref(data, mn)
                    mn *= mult
                    glBufferSubData(GL_ARRAY_BUFFER, mn, sz, ref)
                    dirtyElems[0].clear()
                    dirtyElems[1].clear()
                else:
                    mult = (sizeOf(data) // len(data))
                    glBufferData(GL_ARRAY_BUFFER, len(data) * mult, data, 
                                 GL_STATIC_DRAW)
            GLCommon.bindArray(bindArgs[array])
        self.dirty.clear()