"""
PyGL3Display
(C) David Griffin (2010-present)
"Habilain" is a pseudonym used by David Griffin in a number of places.

A high performance OpenGL library

Licensed under GPLv2 or GPLv3. You should have received a copy of both of these
licenses in the files GPLv2.txt and GPLv3.txt. If not these licenses can be
found at http://www.gnu.org/licenses/gpl-2.0.html (GPLv2) and
http://www.gnu.org/licenses/gpl-3.0.html (GPLv3). 

pygl3display.vertexarrays.spec

Provides the VertexArraySpec objects, which is how a vertexarray object 
should be constructed.
"""

from pygl3display.gl import (GL_FLOAT, GL_FALSE, cdata)
from pygl3display.vertexarrays.obj import VertexArraysOnline
from pygl3display.vertexarrays.accessors import (Tuple, TupleDyn, 
    TupleSingle, TupleSingleDyn, Individual, IndividualDyn, IndividualSingle,
    IndividualSingleDyn, OffsetAccessor)


class VertexArraySpec(object):
    """Defines a specification for a class of vertex arrays"""
    def __init__(self, name, slotMultiplier=1.0):
        """Initialise the VertexArraySpec
        Args:
          name: The name of the class of vertex arrays to create
                (as would be used in documentation)
          slotMultiplier: A multiplier to give an idea of how many render
                          slots are needed per data point. For example, on a
                          sprite 4 data points are needed per sprite, but to
                          draw two triangles 6 render points are needed, so
                          the multiplier is 1.5. This is only a guide, can be
                          ignored if OpenGL wouldn't support a long enough 
                          array, and shouldn't result in too big a performance
                          hit even if it is completely ignored. Default: 1.0
        """
        self.name = name
        self.spec = {}
        self.attribs = []
        self.swappers = []
        self.vertexArraysClass = None
        self.slotMultiplier = slotMultiplier
        
    def copy(self, newName):
        """Return a copy of the vertex arrays spec, but this copy
        will always be unfrozen"""
        copy = type(self)(newName, self.slotMultiplier)
        copy.spec = self.spec.copy()
        copy.attribs = self.attribs[:]
        copy.swappers = self.swappers[:]
        return copy
        
    def addArray(self, name, cType=cdata['f'], glType=GL_FLOAT, 
                       noAttribs=4, glNorm=GL_FALSE, glStride=0,
                       dynamic=True):
        """Add an array to the specification. Defaults to float vec4s"""
        if self.vertexArraysClass is None:
            self.spec[name] = (cType, (noAttribs, glType, glNorm, glStride),
                                dynamic)
        else:
            raise Exception(u'VertexArraySpec has been frozen')
    
    def addAttrib(self, arrayName, attribName, attribSize, attribOffset,
                  attribBatched=False):
        """Add an attribute which will gain an accessor. Accessors
        are accessed by indexing in with the accessor name; accessors
        that take an extra offset parameter must be indexed into again.
        e.g. vertexArrays['name'] for normal, vertexArrays['name'][1]
        for an offset accessor
        Args:
          arrayName: name of the array the accessor indexes
          attribName: name of accessor
          attribSize: size of the attribute
          attribOffset: offset of the attribute; if None, an offset
          parameter will be required by accessor (variable offset)
          attribBatcher: If True, indexing in with a slice will assign
          all values in the index the same value. If False, slice
          assignment works normally."""
        if self.vertexArraysClass is None:
            self.attribs.append((attribName, arrayName, attribSize, 
                                 attribOffset, attribBatched))
        else:
            raise Exception('VertexArraySpec has become frozen')
            
#    def addSwapper(self, name, docstring, swaps):
#        """Add an array swapper to the spec"""
#        if self.vertexArraysClass is None:
#            self.swappers.append((name, docstring, swaps))
#        else:
#            raise Exception('VertexArraySpec has become frozen')
            
    # Index format for setterMaker keys:
    # (attribBatched, singleAttrib, dynamic)
    
    accessors = {
        (True,  True,  False): Individual,
        (True,  False, False): Tuple,
        (False, True,  False): IndividualSingle,
        (False, False, False): TupleSingle,
        (True,  True,  True) : IndividualDyn,
        (True,  False, True) : TupleDyn,
        (False, True,  True) : IndividualSingleDyn,
        (False, False, True) : TupleSingleDyn,
        # These need special creation
    }          
    
    def makeVertexClass(self):
        """Create and metaprogram a class that matches the current spec"""
        class VertexArrayClass(VertexArraysOnline):
            accessorClasses = {}
        self.vertexArraysClass = VertexArrayClass
        VertexArrayClass.__name__ = self.name
        VertexArrayClass.slotMultiplier = self.slotMultiplier
        for attribName, arrayName, attribSize, attribOffset, attribBatched \
        in self.attribs:
            if attribName in VertexArrayClass.accessorClasses:
                raise Exception('Redefined attrib %s' % attribName)
            itemSize = self.spec[arrayName][1][0]
            offsetIsNone = attribOffset is None
            singleAttrib = attribSize == 1
            dynamic = self.spec[arrayName][2]
            signature = (attribBatched, singleAttrib, dynamic)
            AccessorClassBase = self.accessors[signature]
            class AccessorClass(AccessorClassBase):
                """Accessor base class"""
            type(AccessorClass).setup(AccessorClass, arrayName, itemSize, 
                attribSize, (0 if attribOffset is None else attribOffset))
            
            if offsetIsNone is True:
                # Create a special accessor
                class OffsetAccessorClass(OffsetAccessor):
                    """Offset Accessor base class"""
                type(OffsetAccessorClass).setup(OffsetAccessorClass, 
                                                AccessorClass)
                AccessorClass = OffsetAccessorClass
            
            VertexArrayClass.accessorClasses[attribName] = AccessorClass
                                
        #for swapperArgs in self.swappers:
        #    VertexArrayClass.makeSwapper(*swapperArgs)
            
    def makeArrays(self, length):
        """Freeze the spec and return a VertexArrays object of the class
        defined by the spec."""
        if self.vertexArraysClass is None:
            self.makeVertexClass()
        arrays = {}
        for name in self.spec:
            arraySpec = self.spec[name]
            cArray = arraySpec[0](arraySpec[1][0] * length)
            arrays[name] = [cArray, arraySpec[1], arraySpec[2]] 
        return self.vertexArraysClass(length, arrays, None)
